# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Constrained Quadratic Model class.
"""

from __future__ import annotations

import collections.abc as abc
import copy
import json
import re
import tempfile
import uuid
import warnings
import zipfile

from numbers import Number
from typing import Hashable, Optional, Union, BinaryIO, ByteString, Iterable, Collection, Dict
from typing import Callable, MutableMapping, Iterator, Tuple, Mapping, Any, NamedTuple

import numpy as np

from dimod.core.bqm import BQM as BQMabc
from dimod.binary.binary_quadratic_model import BinaryQuadraticModel, Binary, Spin, as_bqm
from dimod.discrete.discrete_quadratic_model import DiscreteQuadraticModel
from dimod.exceptions import InfeasibileModelError
from dimod.quadratic import QuadraticModel
from dimod.sampleset import as_samples
from dimod.sym import Comparison, Eq, Le, Ge, Sense
from dimod.serialization.fileview import SpooledTemporaryFile, _BytesIO
from dimod.serialization.fileview import load, read_header, write_header
from dimod.typing import Bias, Variable
from dimod.utilities import iter_safe_relabels, new_label
from dimod.variables import Variables, serialize_variable, deserialize_variable
from dimod.vartypes import Vartype, as_vartype, VartypeLike
from dimod.serialization.lp import make_lp_grammar, get_variables_from_parsed_lp, constraint_symbols, obj_senses

__all__ = ['ConstrainedQuadraticModel', 'CQM', 'cqm_to_bqm']


CQM_MAGIC_PREFIX = b'DIMODCQM'


class ConstraintData(NamedTuple):
    label: Hashable
    lhs_energy: float
    rhs_energy: float
    sense: Sense
    activity: float
    violation: float


class ConstrainedQuadraticModel:
    r"""A constrained quadratic model.

    Constrained quadratic models are problems of the form:

    .. math::

        \begin{align}
            \text{Minimize an objective:} & \\
            & \sum_{i} a_i x_i + \sum_{i \le j} b_{ij} x_i x_j + c, \\
            \text{Subject to constraints:} & \\
            & \sum_i a_i^{(c)} x_i + \sum_{i \le j} b_{ij}^{(c)} x_i x_j+ c^{(c)} \le 0,
            \quad c=1, \dots, C_{\rm ineq.}, \\
            & \sum_i a_i^{(d)} x_i + \sum_{i \le j} b_{ij}^{(d)} x_i x_j + c^{(d)} = 0,
            \quad d=1, \dots, C_{\rm eq.},
        \end{align}

    where :math:`\{ x_i\}_{i=1, \dots, N}` can be binary\ [#]_ or integer
    variables, :math:`a_{i}, b_{ij}, c` are real values and
    :math:`C_{\rm ineq.}, C_{\rm eq,}` are the number of inequality and
    equality constraints respectively.

    .. [#]
        For binary variables, the range of the quadratic-term summation is
        :math:`i < j` because :math:`x^2 = x` for binary values :math:`\{0, 1\}`
        and :math:`s^2 = 1` for spin values :math:`\{-1, 1\}`.

    The objective and constraints are encoded as either :class:`.QuadraticModel`
    or :class:`.BinaryQuadraticModel` depending on the variable types used.

    Example:

        Solve a simple `bin packing problem <https://w.wiki/3jz4>`_. In this
        problem we wish to pack a set of items of different weights into
        the smallest number of bins possible.

        See :func:`~dimod.generators.bin_packing` for a general function to
        generate bin packing problems. We follow the same naming conventions
        in this example.

        Let's start with four object weights and assume that each bin has a
        capacity of 1.

        >>> weights = [.9, .7, .2, .1]
        >>> capacity = 1

        Let :math:`y_j` indicate that we used bin :math:`j`. We know that we
        will use four or fewer total bins.

        >>> y = [dimod.Binary(f'y_{j}') for j in range(len(weights))]

        Let :math:`x_{i,j}` indicate that we put item :math:`i` in bin
        :math:`j`.

        >>> x = [[dimod.Binary(f'x_{i}_{j}') for j in range(len(weights))]
        ...      for i in range(len(weights))]

        Create an empty constrained quadratic model with no objective or
        constraints.

        >>> cqm = dimod.ConstrainedQuadraticModel()

        We wish to minimize the number of bins used. Therefore our objective
        is to minimize the value of :math:`\sum_j y_j`.

        >>> cqm.set_objective(sum(y))

        We also need to enforce the constraint that each item can only go
        in one bin. We can express this constraint, for a given item :math:`i`,
        with :math:`\sum_j x_{i, j} == 1`. Note that the label of each
        constraint is returned so that we can access them in the future if
        desired.

        >>> for i in range(len(weights)):
        ...     cqm.add_constraint(sum(x[i]) == 1, label=f'item_placing_{i}')
        'item_placing_0'
        'item_placing_1'
        'item_placing_2'
        'item_placing_3'

        Finally, we need to enforce the limits on each bin. We can express
        this constraint, for a given bin :math:`j`, with
        :math:`\sum_i x_{i, j} * w_i <= c` where :math:`w_i` is the weight
        of item :math:`i` and :math:`c` is the capacity.

        >>> for j in range(len(weights)):
        ...     cqm.add_constraint(
        ...         sum(weights[i] * x[i][j] for i in range(len(weights))) - y[j] * capacity <= 0,
        ...         label=f'capacity_bin_{j}')
        'capacity_bin_0'
        'capacity_bin_1'
        'capacity_bin_2'
        'capacity_bin_3'

    """
    def __init__(self):
        # discrete variable tracking, we probably can do this with less memory
        # but for now let's keep it simple
        self.discrete: Set[Hashable] = set()  # collection of discrete constraints

        self._objective = QuadraticModel()

    @property
    def constraints(self) -> Dict[Hashable, Comparison]:
        """The constraints as a dictionary.

        This dictionary and its contents should not be modified.
        """
        try:
            return self._constraints
        except AttributeError:
            pass

        self._constraints: Dict[Hashable, Comparison] = {}
        return self._constraints

    @property
    def objective(self) -> QuadraticModel:
        """The objective to be minimized."""
        return self._objective

    @property
    def variables(self) -> Variables:
        """The variables in use over the objective and all constraints."""
        try:
            return self._variables
        except AttributeError:
            pass

        self._variables = variables = self.objective.variables

        # to support backwards compatibility (0.10.0 - 0.10.5), we annotate
        # this object with some attributes. All of these will be removed in
        # 0.11.0
        def vartype(v):
            warnings.warn(
                "cqm.variables.vartype(v) is deprecated and will be removed in dimod 0.11.0, "
                "use cqm.vartype(v) instead.", DeprecationWarning, stacklevel=2)
            return self.vartype(v)

        variables.vartype = vartype  # method
        variables.vartypes = _Vartypes(self)
        variables.lower_bounds = _LowerBounds(self)
        variables.upper_bounds = _UpperBounds(self)

        return variables

    def _add_variables_from(self, model: Union[BinaryQuadraticModel, QuadraticModel]):
        # todo: singledispatchmethod in 3.8+
        if isinstance(model, (BinaryQuadraticModel, BQMabc)):
            vartype = model.vartype

            for v in model.variables:
                self.objective.add_variable(vartype, v)

        elif isinstance(model, QuadraticModel):
            for v in model.variables:
                # for spin, binary variables the bounds are ignored anyway
                self.objective.add_variable(model.vartype(v), v,
                                            lower_bound=model.lower_bound(v),
                                            upper_bound=model.upper_bound(v))
        else:
            raise TypeError("model should be a QuadraticModel or a BinaryQuadraticModel")

    def add_constraint(self, data, *args, **kwargs) -> Hashable:
        """A convenience wrapper for other methods that add constraints.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Integers
            >>> i, j = Integers(['i', 'j'])
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_constraint(i + j <= 3, label='Constrained i-j range')
            'Constrained i-j range'

        See also:
            :meth:`~.ConstrainedQuadraticModel.add_constraint_from_model`

            :meth:`~.ConstrainedQuadraticModel.add_constraint_from_comparison`

            :meth:`~.ConstrainedQuadraticModel.add_constraint_from_iterable`

        """
        # in python 3.8+ we can use singledispatchmethod
        if isinstance(data, (BinaryQuadraticModel, QuadraticModel, BQMabc)):
            return self.add_constraint_from_model(data, *args, **kwargs)
        elif isinstance(data, Comparison):
            return self.add_constraint_from_comparison(data, *args, **kwargs)
        elif isinstance(data, Iterable):
            return self.add_constraint_from_iterable(data, *args, **kwargs)
        else:
            raise TypeError("unexpected data format")

    def add_constraint_from_model(self,
                                  qm: Union[BinaryQuadraticModel, QuadraticModel],
                                  sense: Union[Sense, str],
                                  rhs: Bias = 0,
                                  label: Optional[Hashable] = None,
                                  copy: bool = True) -> Hashable:
        """Add a constraint from a quadratic model.

        Args:
            qm: A quadratic model or binary quadratic model.

            sense: One of `<=', '>=', '=='.

            rhs: The right hand side of the constraint.

            label: A label for the constraint. Must be unique. If no label
                is provided, then one is generated using :mod:`uuid`.

            copy: If `True`, the BQM is copied. This can be set to `False` to
                improve performance, but subsequently mutating the bqm can
                cause issues.

        Returns:
            The label of the added constraint.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Binary
            >>> cqm = ConstrainedQuadraticModel()
            >>> x = Binary('x')
            >>> cqm.add_constraint_from_model(x, '>=', 0, 'Min x')
            'Min x'
        """
        variables = self.variables

        # get sense as an enum
        if isinstance(sense, str):
            sense = Sense(sense)

        if label is None:
            # we support up to 100k constraints and :6 gives us 16777216
            # possible so pretty safe
            label = uuid.uuid4().hex[:6]
            while label in self.constraints:
                label = uuid.uuid4().hex[:6]
        elif label in self.constraints:
            raise ValueError("a constraint with that label already exists")

        if isinstance(qm, BQMabc):
            qm = as_bqm(qm)  # handle legacy BQMs

        self._add_variables_from(qm)

        if copy:
            qm = qm.copy()

        if sense is Sense.Le:
            self.constraints[label] = Le(qm, rhs)
        elif sense is Sense.Ge:
            self.constraints[label] = Ge(qm, rhs)
        elif sense is Sense.Eq:
            self.constraints[label] = Eq(qm, rhs)
        else:
            raise RuntimeError("unexpected sense")

        return label

    def add_constraint_from_comparison(self,
                                       comp: Comparison,
                                       label: Optional[Hashable] = None,
                                       copy: bool = True) -> Hashable:
        """Add a constraint from a comparison.

        Args:
            comp: A comparison object.

            label: A label for the constraint. Must be unique. If no label
                is provided, one is generated using :mod:`uuid`.

            copy: If `True`, the model is copied. You can set to `False` to
                improve performance, but subsequently mutating the model can
                cause issues.

        Returns:
            Label of the added constraint.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Integer
            >>> i = Integer('i')
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_constraint_from_comparison(i <= 3, label='Max i')
            'Max i'
        """
        if not isinstance(comp.rhs, Number):
            raise TypeError("comparison should have a numeric rhs")

        if isinstance(comp.lhs, (BinaryQuadraticModel, QuadraticModel)):
            return self.add_constraint_from_model(comp.lhs, comp.sense, rhs=comp.rhs,
                                                  label=label, copy=copy)
        else:
            raise ValueError("comparison should have a binary quadratic model "
                             "or quadratic model lhs.")

    def add_constraint_from_iterable(self, iterable: Iterable,
                                     sense: Union[Sense, str],
                                     rhs: Bias = 0,
                                     label: Optional[Hashable] = None,
                                     ) -> Hashable:
        """Add a constraint from an iterable of tuples.

        Args:
            iterable: An iterable of terms as tuples. The variables must
                have already been added to the object.

            sense: One of `<=', '>=', '=='.

            rhs: The right hand side of the constraint.

            label: A label for the constraint. Must be unique. If no label
                is provided, then one is generated using :mod:`uuid`.

        Returns:
            The label of the added constraint.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Integer, Binary
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_variable('i', 'INTEGER')   # doctest: +IGNORE_RESULT
            >>> cqm.add_variable('j', 'INTEGER')   # doctest: +IGNORE_RESULT
            >>> cqm.add_variable('x', 'BINARY')    # doctest: +IGNORE_RESULT
            >>> cqm.add_variable('y', 'BINARY')    # doctest: +IGNORE_RESULT
            >>> label1 = cqm.add_constraint_from_iterable([('x', 'y', 1), ('i', 2), ('j', 3),
            ...                                           ('i', 'j', 1)], '<=', rhs=1)

        """
        qm = self._iterable_to_qm(iterable)

        # use quadratic model in the future
        return self.add_constraint_from_model(
            qm, sense, rhs=rhs, label=label, copy=False)

    def add_discrete(self, data, *args, **kwargs) -> Hashable:
        """Add a one-hot constraint.

        Add a special kind of one-hot constraint. These one-hot constraints
        must be disjoint, that is they must not have any overlapping variables.

        Note that constraints added by :meth:`add_discrete` and other similar
        methods are guaranteed to be satisfied in solutions returned by
        :class:`~dwave.system.samplers.LeapHybridCQMSampler`.

        See also:
            :meth:`~.ConstrainedQuadraticModel.add_discrete_from_model`

            :meth:`~.ConstrainedQuadraticModel.add_discrete_from_comparison`

            :meth:`~.ConstrainedQuadraticModel.add_discrete_from_iterable`

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()

            Add a discrete constraint over variables ``x, y, z`` from an
            iterable.

            >>> iterable = ['x', 'y', 'z']
            >>> for v in iterable:
            ...      cqm.add_variable(v, 'BINARY')
            'x'
            'y'
            'z'
            >>> cqm.add_discrete(iterable, label='discrete-xyz')
            'discrete-xyz'

            Add a discrete constraint over variables ``a, b, c`` from a
            model.

            >>> a, b, c = dimod.Binaries('abc')
            >>> cqm.add_discrete(sum([a, b, c]), label='discrete-abc')
            'discrete-abc'

            Add a discrete constraint over variables ``d, e, f`` from a
            comparison.

            >>> d, e, f = dimod.Binaries('def')
            >>> cqm.add_discrete(d + e + f == 1, label='discrete-def')
            'discrete-def'

        """
        # in python 3.8+ we can use singledispatchmethod
        if isinstance(data, (BinaryQuadraticModel, QuadraticModel, BQMabc)):
            return self.add_discrete_from_model(data, *args, **kwargs)
        elif isinstance(data, Comparison):
            return self.add_discrete_from_comparison(data, *args, **kwargs)
        elif isinstance(data, Iterable):
            return self.add_discrete_from_iterable(data, *args, **kwargs)
        else:
            raise TypeError("unexpected data format")

    def add_discrete_from_comparison(self,
                                     comp: Comparison,
                                     label: Optional[Hashable] = None,
                                     copy: bool = True) -> Hashable:
        """Add a one-hot constraint.

        Add a special kind of one-hot constraint. These one-hot constraints
        must be disjoint, that is they must not have any overlapping variables.

        Note that constraints added by :meth:`add_discrete` and other similar
        methods are guaranteed to be satisfied in solutions returned by
        :class:`~dwave.system.samplers.LeapHybridCQMSampler`.

        Args:
            comp: A comparison object. The comparison must be a linear
                equality constraint with all of the linear biases on the
                left-hand side equal to one and the right hand side equal
                to one.

            label: A label for the constraint. Must be unique. If no label
                is provided, one is generated using :mod:`uuid`.

            copy: If `True`, the model is copied. You can set to `False` to
                improve performance, but subsequently mutating the model can
                cause issues.

        Returns:
            Label of the added constraint.

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> d, e, f = dimod.Binaries('def')
            >>> cqm.add_discrete(d + e + f == 1, label='discrete-def')
            'discrete-def'

        """
        if comp.sense is not Sense.Eq:
            raise ValueError("discrete constraints must be equality constraints")
        if comp.rhs != 1:
            # could scale, but let's keep it simple for now
            raise ValueError("the right hand side of a discrete constraint must be 1")
        return self.add_discrete_from_model(comp.lhs, label=label, copy=copy)

    def add_discrete_from_iterable(self,
                                   variables: Iterable[Variable],
                                   label: Optional[Hashable] = None) -> Hashable:
        """Add a one-hot constraint.

        Add a special kind of one-hot constraint. These one-hot constraints
        must be disjoint, that is they must not have any overlapping variables.

        Note that constraints added by :meth:`add_discrete` and other similar
        methods are guaranteed to be satisfied in solutions returned by
        :class:`~dwave.system.samplers.LeapHybridCQMSampler`.

        Args:
            variables: An iterable of variables.

            label: Label for the constraint. Must be unique. If no label
                is provided, then one is generated using :mod:`uuid`.

        Returns:
            Label of the added constraint.

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> iterable = ['x', 'y', 'z']
            >>> for v in iterable:
            ...      cqm.add_variable(v, 'BINARY')
            'x'
            'y'
            'z'
            >>> cqm.add_discrete(iterable, label='discrete-xyz')
            'discrete-xyz'

        """
        if label is not None and label in self.constraints:
            raise ValueError("a constraint with that label already exists")

        bqm = BinaryQuadraticModel(Vartype.BINARY, dtype=np.float32)

        for v in variables:
            if v in self.variables:
                # it already exists, let's make sure it's not already used
                if any(v in self.constraints[label].lhs.variables for label in self.discrete):
                    raise ValueError(f"variable {v!r} is already used in a discrete variable")
                if self.vartype(v) is not Vartype.BINARY:
                    raise ValueError(f"variable {v!r} has already been added but is not BINARY")

            bqm.set_linear(v, 1)

        if bqm.num_variables < 2:
            raise ValueError("discrete constraints must have at least two variables")

        label = self.add_constraint_from_comparison(bqm == 1, label=label, copy=False)
        self.discrete.add(label)
        return label

    def add_discrete_from_model(self,
                                qm: Union[BinaryQuadraticModel, QuadraticModel],
                                label: Optional[Hashable] = None,
                                copy: bool = True) -> Hashable:
        """Add a one-hot constraint.

        Add a special kind of one-hot constraint. These one-hot constraints
        must be disjoint, that is they must not have any overlapping variables.

        Note that constraints added by :meth:`add_discrete` and other similar
        methods are guaranteed to be satisfied in solutions returned by
        :class:`~dwave.system.samplers.LeapHybridCQMSampler`.

        Args:
            qm: A quadratic model or binary quadratic model.
                The model must be linear with all of the linear biases on the
                left-hand side equal to one and the right hand side equal
                to one.

            label: A label for the constraint. Must be unique. If no label
                is provided, one is generated using :mod:`uuid`.

            copy: If `True`, the model is copied. You can set to `False` to
                improve performance, but subsequently mutating the model can
                cause issues.

        Returns:
            Label of the added constraint.

        Examples:

            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> a, b, c = dimod.Binaries('abc')
            >>> cqm.add_discrete(sum([a, b, c]), label='discrete-abc')
            'discrete-abc'

        """
        vartype = qm.vartype if isinstance(qm, QuadraticModel) else lambda v: qm.vartype

        if qm.num_interactions:
            raise ValueError("discrete constraints must be linear")

        if qm.num_variables < 2:
            raise ValueError("discrete constraints must have at least two variables")

        for v, bias in qm.iter_linear():
            if not vartype(v) is Vartype.BINARY:
                raise ValueError("all variables in a discrete constraint must be binary, "
                                 f"{v!r} is {vartype(v).name!r}")
            # we could maybe do a scaling, but let's just keep it simple for now
            if bias != 1:
                raise ValueError("all linear biases in a discrete constraint must be 1")

        label = self.add_constraint_from_model(qm, sense='==', rhs=1, label=label, copy=copy)
        self.discrete.add(label)
        return label

    def add_variable(self, v: Variable, vartype: VartypeLike,
                     *,
                     lower_bound: Optional[float] = None,
                     upper_bound: Optional[float] = None,
                     ) -> Variable:
        """Add a variable to the model.

        Args:
            variable: A variable label.

            vartype:
                Variable type. One of:

                * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``
                * :class:`.Vartype.INTEGER`, ``'INTEGER'``

            lower_bound:
                A lower bound on the variable. Ignored when the variable is
                not :class:`Vartype.INTEGER`.

            upper_bound:
                An upper bound on the variable. Ignored when the variable is
                not :class:`Vartype.INTEGER`.

        Returns:
            The variable label.

        Raises:
            ValueError: If ``v`` is already a variable in the model and
                the ``vartype``, ``lower_bound``, or ``upper_bound`` are
                inconsistent.

        Examples:
            >>> from dimod import ConstrainedQuadraticModel, Integer
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_variable('i', 'INTEGER')
            'i'

        """
        return self.objective.add_variable(
            vartype, v, lower_bound=lower_bound, upper_bound=upper_bound)

    def check_feasible(self, sample_like, rtol: float = 1e-6, atol: float = 1e-8) -> bool:
        r"""Return the feasibility of the given sample.

        A sample is feasible if all constraints are satisfied. A constraint's
        satisfaction is tested using the following equation:

        .. math::

            violation <= (atol + rtol * | rhs\_energy | )

        where ``violation`` and ``rhs_energy`` are as returned by :meth:`.iter_constraint_data`.

        Args:
            sample_like: A sample.
            rtol: The relative tolerance.
            atol: the absolute tolerance.

        Returns:
            True if the sample is feasible (given the tolerances).

        """
        return all(datum.violation <= atol + rtol*abs(datum.rhs_energy)
                   for datum in self.iter_constraint_data(sample_like))

    def fix_variable(self, v: Variable, value: float, *,
                     cascade: bool = False,
                     ) -> Dict[Variable, float]:
        """Fix the value of a variable in the model.

        Note that this function does not test feasibility.

        Args:
            v: A variable in the model.

            value: A value to assign that variable.

            cascade: If ``True``, then some other variable may be removed
                from the model based on the assignment of ``v``.
                Currently handles the following cases: fixing the value in a
                discrete constraint (see :meth:`add_discrete`) to ``1``; fixing
                all but one of the value in a discrete constratint to `0`;
                fixing a variable in an equality constraint with two
                variables, e.g. fixing ``i`` to ``3`` in ``i + j == 7`` will
                also fix ``j`` to ``4``.

        Returns:
            The assignments of any additional variables fixed.
            If ``cascade==False`` then this will be exactly ``{}``.
            If ``casecade==True`` then additional variable may be fixed.
            See above.

        Raises:
            ValueError: If ``v`` is not a variable in the model.

        """
        if v not in self.variables:
            raise ValueError(f"unknown variable {v!r}")

        new: Dict[Variable, float] = {}
        for label, comparison in self.constraints.items():
            if v in comparison.lhs.variables:
                comparison.lhs.fix_variable(v, value)

                if label in self.discrete:
                    # we've fixed a variable in a discrete constraint, so it
                    # might not be one any more.

                    if value not in Vartype.BINARY.value:
                        raise ValueError(
                            f"variable {v} is part of a discrete constraint so "
                            "it can only be fixed to 0 or 1")

                    if value == 1:  # we've found the one-hot!
                        self.discrete.remove(label)

                        if cascade:
                            # everything else gets set to 0
                            new.update((v, 0) for v in comparison.lhs.variables)

                    elif len(comparison.lhs.variables) < 2:  # we've made it too small
                        self.discrete.remove(label)

                        if cascade and len(comparison.lhs.variables) == 1:
                            new[comparison.lhs.variables[0]] = 1

                elif cascade and comparison.sense is Sense.Eq and len(comparison.lhs.variables) == 1:
                    # we have a constraint like i == 7, so can just go ahead and set it
                    new[comparison.lhs.variables[0]] = comparison.rhs - comparison.lhs.offset

        self.objective.fix_variable(v, value)

        for assignment in new.items():
            self.fix_variable(*assignment)

        return new

    def fix_variables(self,
                      fixed: Union[Mapping[Variable, float], Iterable[Tuple[Variable, float]]],
                      *,
                      cascade: bool = False,
                      ) -> Dict[Variable, float]:
        """Fix the value of the variables and remove them.

        Args:
            fixed: A dictionary or an iterable of 2-tuples of variable assignments.
            cascade: See :meth:`.fix_variable`.

        Returns:
            The assignments of any additional variables fixed.
            If ``cascade==False`` then this will be exactly ``{}``.
            If ``casecade==True`` then additional variable may be fixed.
            See :meth:`.fix_variable`.

        Raises:
            ValueError: If given a variable not in the model.

            :exc:`~dimod.exceptions.InfeasibileModelError`: If fixing the
                given variables will result in an infeasible model. Note that
                this exception is only raised in some obvious cases, it
                is possible to fix variable to create an infeasible model
                without raising this error.

        """
        if isinstance(fixed, Mapping):
            fixed = fixed.items()

        fix_variable = self.fix_variable

        new: Dict[Variable, float] = {}
        for v, val in fixed:
            if v in new:
                if new[v] != val:
                    raise InfeasibileModelError()
            else:
                new.update(fix_variable(v, val, cascade=cascade))

        return new

    def flip_variable(self, v: Variable):
        """Flip the specified variable in the objective and constraints.

        Note that this may remove a constraint as being tracked as discrete.
        Subsequently flipping the variable again will not restore it.
        """
        self.objective.flip_variable(v)  # checks that it exists and is the correct vartype

        for label, comparison in self.constraints.items():
            lhs = comparison.lhs
            if v in lhs.variables:
                comparison.lhs.flip_variable(v)

                self.discrete.discard(label)  # no longer a discrete variable

    @classmethod
    def from_bqm(cls, bqm: BinaryQuadraticModel) -> 'ConstrainedQuadraticModel':
        """Alias for :meth:`from_quadratic_model`."""
        return cls.from_quadratic_model(bqm)

    @classmethod
    def from_discrete_quadratic_model(cls, dqm: DiscreteQuadraticModel, *,
                                      relabel_func: Callable[[Variable, int], Variable] = lambda v, c: (v, c),
                                      ) -> 'ConstrainedQuadraticModel':
        """Construct a constrained quadratic model from a discrete quadratic model.

        Args:
            dqm: a discrete quadratic model.

            relabel_func (optional): A function that takes two arguments, the
                variable label and the case label, and returns a new variable
                label to be used in the CQM. By default generates a 2-tuple
                `(variable, case)`.

        Returns:
            A constrained quadratic model.

        """
        cqm = cls()

        objective = BinaryQuadraticModel(Vartype.BINARY)

        seen = set()
        for v in dqm.variables:
            seen.add(v)

            # convert v, case to a flat set of variables
            v_vars = list(relabel_func(v, case) for case in dqm.get_cases(v))

            # add the one-hot constraint
            cqm.add_discrete(v_vars, label=v)

            # add to the objective
            objective.add_linear_from(zip(v_vars, dqm.get_linear(v)))

            for u in dqm.adj[v]:
                if u in seen:  # only want upper-triangle
                    continue

                u_vars = list(relabel_func(u, case) for case in dqm.get_cases(u))

                objective.add_quadratic_from(
                    (u_vars[cu], v_vars[cv], bias)
                    for (cu, cv), bias
                    in dqm.get_quadratic(u, v).items()
                    )

        objective.offset = dqm.offset

        cqm.set_objective(objective)

        return cqm

    from_dqm = from_discrete_quadratic_model

    @classmethod
    def from_quadratic_model(cls, qm: Union[QuadraticModel, BinaryQuadraticModel]
                             ) -> 'ConstrainedQuadraticModel':
        """Construct a constrained quadratic model from a quadratic model or
        binary quadratic model.

        The specified model is set as the objective to be minimzed in the constructed
        constrained quadratic model (CQM). You can then add constraints that any feasible
        solutions should meet.

        Args:
            qm: Binary quadratic model (BQM) or quadratic model (QM).

        Examples:
            This example creates a CQM to minimize a triangular problem with the added
            constraint that one of the variables must have value 1 in feasible solutions.

            >>> from dimod import ConstrainedQuadraticModel, BinaryQuadraticModel
            >>> bqm = BinaryQuadraticModel.from_ising({}, {'ab': 1, 'bc': 1, 'ac': 1})
            >>> cqm = ConstrainedQuadraticModel().from_bqm(bqm)
            >>> cqm.objective.linear
            {'a': 0.0, 'b': 0.0, 'c': 0.0}
            >>> cqm.objective.quadratic
            {('b', 'a'): 1.0, ('c', 'a'): 1.0, ('c', 'b'): 1.0}
            >>> label1 = cqm.add_constraint_from_model(BinaryQuadraticModel({'a': 0}, {}, 0, 'SPIN'), '>=', 0)
        """
        cqm = cls()
        cqm.set_objective(qm)
        return cqm

    @classmethod
    def from_qm(cls, qm: QuadraticModel) -> 'ConstrainedQuadraticModel':
        """Alias for :meth:`from_quadratic_model`."""
        return cls.from_quadratic_model(qm)

    @classmethod
    def from_file(cls, fp: Union[BinaryIO, ByteString]) -> "ConstrainedQuadraticModel":
        """Construct from a file-like object.

        The inverse of :meth:`~ConstrainedQuadraticModel.to_file`.
        """
        if isinstance(fp, ByteString):
            file_like: BinaryIO = _BytesIO(fp)  # type: ignore[assignment]
        else:
            file_like = fp

        header_info = read_header(file_like, CQM_MAGIC_PREFIX)

        if header_info.version >= (2, 0):
            raise ValueError("cannot load a BQM serialized with version "
                             f"{header_info.version!r}, try upgrading your "
                             "dimod version")

        # we don't actually need the data

        cqm = CQM()

        with zipfile.ZipFile(file_like, mode='r') as zf:
            cqm.set_objective(load(zf.read("objective")))

            constraint_labels = set()
            for arch in zf.namelist():
                # even on windows zip uses /
                match = re.match("constraints/([^/]+)/", arch)
                if match is not None:
                    constraint_labels.add(match.group(1))

            for constraint in constraint_labels:
                lhs = load(zf.read(f"constraints/{constraint}/lhs"))
                rhs = np.frombuffer(zf.read(f"constraints/{constraint}/rhs"), np.float64)[0]
                sense = zf.read(f"constraints/{constraint}/sense").decode('ascii')
                discrete = any(zf.read(f"constraints/{constraint}/discrete"))
                label = deserialize_variable(json.loads(constraint))
                cqm.add_constraint(lhs, rhs=rhs, sense=sense, label=label)
                if discrete:
                    cqm.discrete.add(label)

        return cqm

    def iter_constraint_data(self, sample_like) -> Iterator[ConstraintData]:
        """Yield information about the constraints for the given sample.

        Args:
            sample_like: A sample.

        Yields:
            A :class:`collections.namedtuple` with ``label``, ``lhs_energy``,
            ``rhs_energy``, ``sense``, ``activity``, and ``violation`` fields.
            ``label`` is the constraint label.
            ``lhs_energy`` is the energy of the left hand side of the constraint.
            ``rhs_energy`` is the energy of the right hand side of the constraint.
            ``sense`` is the :class:`dimod.sym.Sense` of the constraint.
            ``activity`` is ``lhs_energy - rhs_energy``
            ``violation`` is determined by the type of constraint. If ``violation``
            is positive, that means that the constraint has been violated by
            that amount. If it is negative, that means that the constraint has
            been satisfied by the amount.

        """

        sample, labels = as_samples(sample_like)

        if sample.shape[0] != 1:
            raise ValueError("sample_like should be a single sample, "
                             f"received {sample.shape[0]} samples")

        for label, constraint in self.constraints.items():
            lhs = constraint.lhs.energy((sample, labels))
            rhs = constraint.rhs
            sense = constraint.sense

            activity = lhs - rhs

            if sense is Sense.Eq:
                violation = abs(activity)
            elif sense is Sense.Ge:
                violation = -activity
            elif sense is Sense.Le:
                violation = activity
            else:
                raise RuntimeError("unexpected sense")

            yield ConstraintData(
                activity=activity,
                sense=sense,
                violation=violation,
                lhs_energy=lhs,
                rhs_energy=rhs,
                label=label,
                )

    def iter_violations(self, sample_like, *, skip_satisfied: bool = False, clip: bool = False,
                        ) -> Iterator[Tuple[Hashable, Bias]]:
        """Yield violations for all constraints.

        Args:
            sample_like: A sample over the CQM variables.
            skip_satisfied: If True, does not yield constraints that are satisfied.
            clip: If True, negative violations are rounded up to 0.

        Yields:
            A 2-tuple containing the constraint label and the amount of
            constraints violation.

        Example:

            Construct a constrained quadratic model.

            >>> i, j, k = dimod.Binaries(['i', 'j', 'k'])
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint(i + j + k == 10, label='equal')
            'equal'
            >>> cqm.add_constraint(i + j <= 15, label='less equal')
            'less equal'
            >>> cqm.add_constraint(j - k >= 0, label='greater equal')
            'greater equal'

            Check the violations of a sample that satisfies all constraints.

            >>> sample = {'i': 3, 'j': 5, 'k': 2}
            >>> for label, violation in cqm.iter_violations(sample, clip=True):
            ...     print(label, violation)
            equal 0.0
            less equal 0.0
            greater equal 0.0

            Check the violations for a sample that does not satisfy all of the
            constraints.

            >>> sample = {'i': 3, 'j': 2, 'k': 5}
            >>> for label, violation in cqm.iter_violations(sample, clip=True):
            ...     print(label, violation)
            equal 0.0
            less equal 0.0
            greater equal 3.0

            >>> sample = {'i': 3, 'j': 2, 'k': 5}
            >>> for label, violation in cqm.iter_violations(sample, skip_satisfied=True):
            ...     print(label, violation)
            greater equal 3.0

        """
        if skip_satisfied:
            # clip doesn't matter in this case
            # todo: feasibility tolerance?
            for datum in self.iter_constraint_data(sample_like):
                if datum.violation > 0:
                    yield datum.label, datum.violation
        elif clip:
            for datum in self.iter_constraint_data(sample_like):
                yield datum.label, max(datum.violation, 0.0)
        else:
            for datum in self.iter_constraint_data(sample_like):
                yield datum.label, datum.violation

    def is_almost_equal(self, other: 'ConstrainedQuadraticModel',
                        places: int = 7) -> bool:
        """Return True if the given model's objective and constraints are almost equal."""
        def constraint_eq(c0: Comparison, c1: Comparison) -> bool:
            return (c0.sense is c1.sense
                    and c0.lhs.is_almost_equal(c1.lhs, places=places)
                    and not round(c0.rhs - c1.rhs, places))

        return (self.objective.is_almost_equal(other.objective, places=places)
                and self.constraints.keys() == other.constraints.keys()
                and all(constraint_eq(constraint, other.constraints[label])
                        for label, constraint in self.constraints.items()))

    def is_equal(self, other: 'ConstrainedQuadraticModel') -> bool:
        """Return True if the given model has the same objective and constraints."""
        def constraint_eq(c0: Comparison, c1: Comparison) -> bool:
            return (c0.sense is c1.sense
                    and c0.lhs.is_equal(c1.lhs)
                    and c0.rhs == c1.rhs)

        return (self.objective.is_equal(other.objective)
                and self.constraints.keys() == other.constraints.keys()
                and all(constraint_eq(constraint, other.constraints[label])
                        for label, constraint in self.constraints.items()))

    def lower_bound(self, v: Variable) -> Bias:
        """Return the lower bound on the specified variable."""
        return self.objective.lower_bound(v)

    def num_biases(self) -> int:
        """The number of biases accross the objective and constraints."""
        num_biases = len(self.objective.linear) + len(self.objective.quadratic)
        num_biases += sum(len(const.lhs.linear) + len(const.lhs.quadratic)
                          for const in self.constraints.values())
        return num_biases

    def num_quadratic_variables(self) -> int:
        """Return the total number of variables with at least one quadratic
        interaction accross all constraints."""
        count = 0
        for const in self.constraints.values():
            lhs = const.lhs
            count += sum(lhs.degree(v) > 0 for v in lhs.variables)
        return count

    def relabel_constraints(self, mapping: Mapping[Hashable, Hashable]):
        """Relabel the constraints.

        Note that this method does not maintain the constraint order.

        Args:
            mapping: A mapping from the old constraint labels to the new.

        Examples:
            >>> x, y, z = dimod.Binaries(['x', 'y', 'z'])
            >>> cqm = dimod.ConstrainedQuadraticModel()
            >>> cqm.add_constraint(x + y == 1, label='c0')
            'c0'
            >>> cqm.add_constraint(y*z == 0, label='c1')
            'c1'
            >>> cqm.relabel_constraints({'c1': 'c2'})
            >>> list(cqm.constraints)
            ['c0', 'c2']

        """
        for submap in iter_safe_relabels(mapping, self.constraints):
            for old, new in submap.items():
                try:
                    self.constraints[new] = self.constraints[old]
                except KeyError:
                    continue  # do nothing
                del self.constraints[old]

                if old in self.discrete:
                    self.discrete.add(new)
                    self.discrete.remove(old)

    def relabel_variables(self,
                          mapping: Mapping[Variable, Variable],
                          inplace: bool = True,
                          ) -> 'ConstrainedQuadraticModel':
        """Relabel the variables of the objective and constraints.

        Args:
            mapping: A mapping from the old variable labels to the new.
            inplace: Whether to act on the model in-place or return a copy.

        Returns:
            A constrained quadratic model. Either ``self`` if ``inplace`` or a
            copy.

        """
        if not inplace:
            return copy.deepcopy(self).relabel_variables(mapping, inplace=True)

        self.objective.relabel_variables(mapping, inplace=True)
        for constraint in self.constraints.values():
            constraint.lhs.relabel_variables(mapping, inplace=True)

        return self

    def remove_constraint(self, label: Hashable, *, cascade: bool = False):
        """Remove a constraint from the model.

        Args:
            label: The constraint label.
            cascade: If ``cascade`` is true, then any variables in the removed
                constraint that are not in any other constraints and that
                contribute no energy to the objective will also be removed.

        """
        try:
            comparison = self.constraints.pop(label)
        except KeyError:
            raise ValueError(f"{label!r} is not a constraint") from None
        self.discrete.discard(label)  # if it's discrete

        if cascade:
            for v in comparison.lhs.variables:
                if self.objective.degree(v) or self.objective.get_linear(v):
                    # it's used somewhere in the objective
                    continue

                if any(v in comp.lhs.variables for comp in self.constraints.values()):
                    # it's used in at least one constraint
                    continue

                self.objective.remove_variable(v)

    def set_lower_bound(self, v: Variable, lb: float):
        """Set the lower bound for a variable.

        Args:
            v: Variable in the constrained quadratic model.
            lb: Lower bound to set for variable ``v``.

        Raises:
            ValueError: If ``v`` is a :class:`~dimod.Vartype.SPIN`
                or :class:`~dimod.Vartype.BINARY` variable.

        """
        self.objective.set_lower_bound(v, lb)
        for comp in self.constraints.values():
            qm = comp.lhs
            if v in qm.variables:
                qm.set_lower_bound(v, lb)

    def set_objective(self, objective: Union[BinaryQuadraticModel,
                                             QuadraticModel, Iterable]):
        """Set the objective of the constrained quadratic model.

        Args:
            objective: Binary quadratic model (BQM) or quadratic model (QM) or
            an iterable of tuples.

        Examples:
            >>> from dimod import Integer, ConstrainedQuadraticModel
            >>> i = Integer('i')
            >>> j = Integer('j')
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.set_objective(2*i - 0.5*i*j + 10)

        """
        if isinstance(objective, Iterable):
            objective = self._iterable_to_qm(objective)
        # clear out current objective, keeping only the variables
        self.objective.quadratic.clear()  # there may be a more performant way...
        for v in self.objective.variables:
            self.objective.set_linear(v, 0)
        # offset is overwritten later

        # now add everything from the new objective
        self._add_variables_from(objective)

        for v in objective.variables:
            self.objective.set_linear(v, objective.get_linear(v))
        self.objective.add_quadratic_from(objective.iter_quadratic())
        self.objective.offset = objective.offset

    def set_upper_bound(self, v: Variable, ub: float):
        """Set the upper bound for a variable.

        Args:
            v: Variable in the constrained quadratic model.
            ub: Upper bound to set for variable ``v``.

        Raises:
            ValueError: If ``v`` is a :class:`~dimod.Vartype.SPIN`
                or :class:`~dimod.Vartype.BINARY` variable.

        """
        self.objective.set_upper_bound(v, ub)
        for comp in self.constraints.values():
            qm = comp.lhs
            if v in qm.variables:
                qm.set_upper_bound(v, ub)

    def _iterable_to_qm(self, iterable: Iterable) -> QuadraticModel:
        qm = QuadraticModel()

        def _add_variable(v):
            # handles vartype, and bounds
            vartype = self.vartype(v)

            if vartype is not Vartype.SPIN and vartype is not Vartype.BINARY:
                # need to worry about bounds
                qm.add_variable(vartype, v,
                                lower_bound=self.lower_bound(v),
                                upper_bound=self.upper_bound(v))
            else:
                qm.add_variable(vartype, v)

        for *variables, bias in iterable:
            if len(variables) == 0:
                qm.offset += bias
            elif len(variables) == 1:
                v, = variables
                _add_variable(v)
                qm.add_linear(v, bias)
            elif len(variables) == 2:
                u, v = variables
                _add_variable(u)
                _add_variable(v)
                qm.add_quadratic(u, v, bias)
            else:
                raise ValueError("terms must be constant, linear or quadratic")
        return qm

    def _substitute_self_loops_from_model(self, qm: Union[BinaryQuadraticModel, QuadraticModel],
                                          mapping: MutableMapping[Variable, Variable]):
        if isinstance(qm, BinaryQuadraticModel):
            # bqms never have self-loops
            return

        for u in qm.variables:
            vartype = qm.vartype(u)

            # integer and binary variables never have self-loops
            if vartype is Vartype.SPIN or vartype is Vartype.BINARY:
                continue

            try:
                bias = qm.get_quadratic(u, u)
            except ValueError:
                # no self-loop
                continue

            lb = qm.lower_bound(u)
            ub = qm.upper_bound(u)

            if u not in mapping:
                # we've never seen this integer before
                new: Variable = new_label()

                # on the off chance there are conflicts. Luckily self.variables
                # is global accross all constraints/objective so we don't need
                # to worry about accidentally picking something we'll regret
                while new in self.constraints or new in self.variables:
                    new = new_label()

                mapping[u] = new

                self.objective.add_variable(vartype, new, lower_bound=lb, upper_bound=ub)

                # we don't add the constraint yet because we don't want
                # to modify self.constraints
            else:
                new = mapping[u]

            qm.add_variable(vartype, new, lower_bound=lb, upper_bound=ub)

            qm.add_quadratic(u, new, bias)
            qm.remove_interaction(u, u)

    def substitute_self_loops(self) -> Dict[Variable, Variable]:
        """Replace any integer self-loops in the objective or constraints.

        Self-loop :math:`i^2` is removed by introducing a new variable
        :math:`j` with interaction :math:`i*j` and adding constraint
        :math:`j == i`.

        Acts on the objective and constraints in-place.

        Returns:
            Mapping from the integer variable labels to their introduced
            counterparts. The constraint enforcing :math:`j == i` uses
            the same label.

        Examples:
            >>> from dimod import Integer, ConstrainedQuadraticModel
            >>> i = Integer('i')
            >>> cqm = ConstrainedQuadraticModel()
            >>> cqm.add_constraint(i*i <=3, label='i squared')
            'i squared'
            >>> cqm.substitute_self_loops()                      # doctest: +IGNORE_RESULT
            >>> cqm.constraints   # doctest: +IGNORE_RESULT
            {'i squared': QuadraticModel({'i': 0.0, 'cf651f3d-bdf8-4735-9139-eee0a32e217f': 0.0}, {('cf651f3d-bdf8-4735-9139-eee0a32e217f', 'i'): 1.0}, 0.0, {'i': 'INTEGER', 'cf651f3d-bdf8-4735-9139-eee0a32e217f': 'INTEGER'}, dtype='float64') <= 3,
            'cf651f3d-bdf8-4735-9139-eee0a32e217f': QuadraticModel({'i': 1.0, 'cf651f3d-bdf8-4735-9139-eee0a32e217f': -1.0}, {}, 0.0, {'i': 'INTEGER', 'cf651f3d-bdf8-4735-9139-eee0a32e217f': 'INTEGER'}, dtype='float64') == 0}
        """
        mapping: Dict[Variable, Variable] = dict()

        self._substitute_self_loops_from_model(self.objective, mapping)

        for comparison in self.constraints.values():
            self._substitute_self_loops_from_model(comparison.lhs, mapping)

        # finally add the constraints for the variables
        for v, new in mapping.items():
            self.add_constraint([(v, 1), (new, -1)], rhs=0, sense='==', label=new)

        return mapping

    def to_file(self, *, spool_size: int = int(1e9)) -> tempfile.SpooledTemporaryFile:
        """Serialize to a file-like object.

        Args:
            spool_size: Defines the `max_size` passed to the constructor of
                :class:`tempfile.SpooledTemporaryFile`. Determines whether
                the returned file-like's contents will be kept on disk or in
                memory.

        Format Specification (Version 1.1):

            This format is inspired by the `NPY format`_

            The first 8 bytes are a magic string: exactly "DIMODCQM".

            The next 1 byte is an unsigned byte: the major version of the file
            format.

            The next 1 byte is an unsigned byte: the minor version of the file
            format.

            The next 4 bytes form a little-endian unsigned int, the length of
            the header data HEADER_LEN.

            The next HEADER_LEN bytes form the header data. This is a
            json-serialized dictionary. The dictionary is exactly:

            .. code-block:: python

                dict(num_variables=len(cqm.variables),
                     num_constraints=len(cqm.constraints),
                     num_biases=cqm.num_biases(),
                     num_quadratic_variables=cqm.num_quadratic_variables(),
                     )

            it is terminated by a newline character and padded with spaces to
            make the entire length of the entire header divisible by 64.

            The constraint quadratic model data comes after the header. It is
            encoded as a zip file. The zip file will contain one file
            named `objective`, containing the objective as encoded as a file
            view. It will also contain a directory called `constraints`. The
            `constraints` directory will contain one subdirectory for each
            constraint, each containing `lhs`, `rhs` and `sense` encoding
            the `lhs` as a fileview, the `rhs` as a float and the sense
            as a string. Each directory will also contain a `discrete` file,
            encoding whether the constraint represents a discrete variable.

        Format Specification (Version 1.0):

            This format is the same as Version 1.1, except that the data dict
            does not have `num_quadratic_variables`.

        .. _NPY format: https://numpy.org/doc/stable/reference/generated/numpy.lib.format.html

        """
        file = SpooledTemporaryFile(max_size=spool_size)

        data = dict(num_variables=len(self.variables),
                    num_constraints=len(self.constraints),
                    num_biases=self.num_biases(),
                    num_quadratic_variables=self.num_quadratic_variables(),
                    )

        write_header(file, CQM_MAGIC_PREFIX, data, version=(1, 1))

        # write the values
        with zipfile.ZipFile(file, mode='a') as zf:
            try:
                zf.writestr(
                    'objective', self.objective.to_file(spool_size=int(1e12))._file.getbuffer())
            except AttributeError:
                # no objective to write
                pass

            for label, constraint in self.constraints.items():
                # put everything in a constraints/label/ directory
                lstr = json.dumps(serialize_variable(label))

                lhs = constraint.lhs.to_file(spool_size=int(1e12))._file.getbuffer()
                zf.writestr(f'constraints/{lstr}/lhs', lhs)

                rhs = np.float64(constraint.rhs).tobytes()
                zf.writestr(f'constraints/{lstr}/rhs', rhs)

                sense = bytes(constraint.sense.value, 'ascii')
                zf.writestr(f'constraints/{lstr}/sense', sense)

                discrete = bytes((label in self.discrete,))
                zf.writestr(f'constraints/{lstr}/discrete', discrete)

        file.seek(0)
        return file

    def upper_bound(self, v: Variable) -> Bias:
        """Return the upper bound on the specified variable."""
        return self.objective.upper_bound(v)

    def vartype(self, v: Variable) -> Vartype:
        """The vartype of the given variable."""
        return self.objective.vartype(v)

    def violations(self, sample_like, *, skip_satisfied: bool = False, clip: bool = False,
                   ) -> Dict[Hashable, Bias]:
        """Return a dictionary mapping constraint labels to the amount the constraints are violated.

        This method is a shortcut for ``dict(cqm.iter_violations(sample))``.
        """
        return dict(self.iter_violations(sample_like, skip_satisfied=skip_satisfied, clip=clip))

    @classmethod
    def from_lp_file(cls,
                     fp: Union[BinaryIO, ByteString],
                     lower_bound_default: Optional[int] = None,
                     upper_bound_default: Optional[int] = None) -> "ConstrainedQuadraticModel":
        """Create a constrained quadratic model from an LP file.

        Args:
            fp: file-like object in LP (linear program) format.
            lower_bound_default: Default lower bound for integer variables.
            upper_bound_default: Default upper bound for integer variables.

        Returns:
            :class:`ConstrainedQuadraticModel` representing the model encoded in
            the LP file.
        """
        grammar = make_lp_grammar()
        parse_output = grammar.parseFile(fp)
        obj = get_variables_from_parsed_lp(parse_output, lower_bound_default, upper_bound_default)

        cqm = ConstrainedQuadraticModel()

        # parse and set the objective
        obj_coefficient = 1
        for oe in parse_output.objective:

            if isinstance(oe, str):
                if oe in obj_senses.keys():
                    obj_coefficient = obj_senses[oe]

            else:
                if len(oe) == 2:
                    if oe.name != "":
                        # linear term
                        obj.add_linear(oe.name[0], obj_coefficient * oe.coef)

                    else:
                        # this is a pure squared term
                        varname = oe.squared_name[0]
                        vartype = obj.vartype(varname)

                        if vartype is Vartype.BINARY:
                            obj.add_linear(varname, obj_coefficient * oe.coef * 0.5)
                        elif vartype is Vartype.INTEGER:
                            obj.add_quadratic(varname, varname,
                                              obj_coefficient * oe.coef * 0.5)
                        else:
                            raise TypeError("Unexpected variable type: {}".format(vartype))

                elif len(oe) == 3:
                    # bilinear term
                    var1 = oe.name[0]
                    var2 = oe.second_var_name[0]
                    obj.add_quadratic(var1, var2, obj_coefficient * oe.coef * 0.5)

        cqm.set_objective(obj)

        # adding constraints
        for c in parse_output.constraints:

            try:
                cname = c.name[0]
            except IndexError:
                # the constraint is nameless, set this to None now
                cname = None

            csense = constraint_symbols[c.sense]
            ccoef = c.rhs
            constraint = QuadraticModel()

            if not c.lin_expr and not c.quad_expr:
                # empty constraint
                warnings.warn('The LP file contained an empty constraint and it will be ignored',
                              stacklevel=2)
                continue

            if c.lin_expr:

                for le in c.lin_expr:
                    var = le.name[0]
                    vartype = obj.vartype(var)
                    lb = obj.lower_bound(var)
                    ub = obj.upper_bound(var)
                    if vartype is Vartype.BINARY:
                        constraint.add_variable(Vartype.BINARY, var)
                    elif vartype is Vartype.INTEGER:
                        constraint.add_variable(Vartype.INTEGER, var, lower_bound=lb, upper_bound=ub)
                    else:
                        raise ValueError("Unexpected vartype: {}".format(vartype))
                    constraint.add_linear(var, le.coef)

            if c.quad_expr:

                for qe in c.quad_expr:
                    if qe.name != "":
                        var1 = qe.name[0]
                        vartype1 = obj.vartype(var1)
                        lb1 = obj.lower_bound(var1)
                        ub1 = obj.upper_bound(var1)

                        if vartype1 is Vartype.BINARY:
                            constraint.add_variable(vartype1, var1)
                        elif vartype1 is Vartype.INTEGER:
                            constraint.add_variable(vartype1, var1, lower_bound=lb1, upper_bound=ub1)
                        else:
                            raise ValueError("Unexpected vartype: {}".format(vartype1))

                        var2 = qe.second_var_name[0]
                        vartype2 = obj.vartype(var2)
                        lb2 = obj.lower_bound(var2)
                        ub2 = obj.upper_bound(var2)

                        if vartype2 is Vartype.BINARY:
                            constraint.add_variable(vartype2, var2)
                        elif vartype2 is Vartype.INTEGER:
                            constraint.add_variable(vartype2, var2, lower_bound=lb2, upper_bound=ub2)

                        constraint.add_quadratic(var1, var2, qe.coef)

                    else:
                        # this is a pure squared term
                        var = qe.squared_name[0]
                        vartype = obj.vartype(var)
                        lb = obj.lower_bound(var)
                        ub = obj.upper_bound(var)
                        if vartype is Vartype.BINARY:
                            constraint.add_variable(vartype, var)
                            constraint.add_linear(var, qe.coef)
                        elif vartype is Vartype.INTEGER:
                            constraint.add_variable(vartype, var, lower_bound=lb, upper_bound=ub)
                            constraint.add_quadratic(var, var, qe.coef)
                        else:
                            raise TypeError("Unexpected variable type: {}".format(vartype))

            # finally mode the RHS to the LHS with a minus sign
            constraint.offset = - ccoef

            cqm.add_constraint(constraint, label=cname, sense=csense)

        return cqm


CQM = ConstrainedQuadraticModel


class _Vartypes(abc.Sequence):
    """Support deprecated attribute on ``CQM.variables``"""
    def __init__(self, cqm: ConstrainedQuadraticModel):
        self.cqm: ConstrainedQuadraticModel = cqm

    def __getitem__(self, index: int) -> Vartype:
        warnings.warn(
            "cqm.variables.vartypes[i] is deprecated and will be removed in dimod 0.11.0, "
            "use cqm.vartype(cqm.variables[i]) instead.", DeprecationWarning, stacklevel=3)
        return self.cqm.vartype(self.cqm.variables[index])

    def __len__(self) -> int:
        warnings.warn(
            "cqm.variables.vartypes is deprecated and will be removed in dimod 0.11.0",
            DeprecationWarning, stacklevel=3)
        return len(self.cqm.variables)


class _LowerBounds(abc.Mapping):
    """Support deprecated attribute on ``CQM.variables``"""
    def __init__(self, cqm: ConstrainedQuadraticModel):
        self.cqm: ConstrainedQuadraticModel = cqm

    def __getitem__(self, key: Variable) -> float:
        warnings.warn(
            "cqm.variables.lower_bounds[v] is deprecated and will be removed in dimod 0.11.0, "
            "use cqm.lower_bound(v) instead.", DeprecationWarning, stacklevel=3)
        return self.cqm.lower_bound(key)

    def __iter__(self) -> Iterator[Variable]:
        warnings.warn(
            "cqm.variables.lower_bounds is deprecated and will be removed in dimod 0.11.0",
            DeprecationWarning, stacklevel=3)
        yield from self.cqm.variables

    def __len__(self) -> int:
        warnings.warn(
            "cqm.variables.lower_bounds is deprecated and will be removed in dimod 0.11.0",
            DeprecationWarning, stacklevel=3)
        return len(self.cqm.variables)


class _UpperBounds(abc.Mapping):
    """Support deprecated attribute on ``CQM.variables``"""
    def __init__(self, cqm: ConstrainedQuadraticModel):
        self.cqm: ConstrainedQuadraticModel = cqm

    def __getitem__(self, key: Variable) -> float:
        warnings.warn(
            "cqm.variables.upper_bounds[v] is deprecated and will be removed in dimod 0.11.0, "
            "use cqm.upper_bound(v) instead.", DeprecationWarning, stacklevel=3)
        return self.cqm.upper_bound(key)

    def __iter__(self) -> Iterator[Variable]:
        warnings.warn(
            "cqm.variables.upper_bounds is deprecated and will be removed in dimod 0.11.0",
            DeprecationWarning, stacklevel=3)
        yield from self.cqm.variables

    def __len__(self) -> int:
        warnings.warn(
            "cqm.variables.upper_bounds is deprecated and will be removed in dimod 0.11.0",
            DeprecationWarning, stacklevel=3)
        return len(self.cqm.variables)


def _qm_to_bqm(qm: QuadraticModel, integers: MutableMapping[Variable, BinaryQuadraticModel],
               ) -> BinaryQuadraticModel:
    # dev note: probably we'll want to make this function or something similar
    # public facing at some point, but right now the interface is pretty weird
    # and it only returns BINARY bqms

    if any(qm.vartype(v) is Vartype.SPIN for v in qm.variables):
        # bqm is BINARY so we want to handle these
        qm = qm.spin_to_binary(inplace=False)

    bqm = BinaryQuadraticModel(Vartype.BINARY)

    for v in qm.variables:
        if v in integers:
            bqm += qm.get_linear(v) * integers[v]
        else:
            bqm.add_linear(v, qm.get_linear(v))

    for u, v, bias in qm.iter_quadratic():
        if u in integers:
            if v in integers:
                bqm += integers[u] * integers[v] * bias
            else:
                bqm += Binary(v) * integers[u] * bias
        elif v in integers:
            bqm += Binary(u) * integers[v] * bias
        else:
            bqm.add_quadratic(u, v, bias)
    bqm.offset += qm.offset

    return bqm


class CQMToBQMInverter:
    """Invert a sample from a binary quadratic model constructed by :func:`cqm_to_bqm`."""
    __slots__ = ('_binary', '_integers')

    def __init__(self,
                 binary: Mapping[Variable, Vartype],
                 integers: Mapping[Variable, BinaryQuadraticModel]):
        self._binary = binary
        self._integers = integers

    def __call__(self, sample: Mapping[Variable, int]) -> Mapping[Variable, int]:
        new = {}

        for v, vartype in self._binary.items():
            if vartype is Vartype.BINARY:
                new[v] = sample[v]
            elif vartype is Vartype.SPIN:
                new[v] = 2*sample[v] - 1
            else:
                raise RuntimeError("unexpected vartype")

        for v, bqm in self._integers.items():
            new[v] = 0
            for u in bqm.variables:
                new[v] += sample[u] * u[1]

        return new

    @classmethod
    def from_dict(cls, doc: Dict[str, Dict[Variable, Any]]) -> 'CQMToBQMInverter':
        """Construct an inverter from a serialized representation."""

        integers = {}
        for v, variables in doc['integers'].items():
            v = deserialize_variable(v)

            bqm = BinaryQuadraticModel(Vartype.BINARY)
            bqm.add_linear_from((deserialize_variable(u), u[1]) for u in variables)

            integers[v] = bqm

        return cls(
            dict((deserialize_variable(v), as_vartype(vartype))
                 for v, vartype in doc['binary'].items()),
            integers,
            )

    def to_dict(self) -> Dict[str, Dict[Variable, Any]]:
        """Return a json-serializable encoding of the inverter."""
        # todo: in 3.8 we can used TypedDict for the typing
        return dict(
            binary=dict((serialize_variable(v), vartype.name)
                        for v, vartype in self._binary.items()),
            integers=dict((serialize_variable(v), bqm.variables.to_serializable())
                          for v, bqm in self._integers.items()),
            )


# Developer note: This function is *super* ad hoc. In the future, we may want
# A BQM.from_cqm method or similar, but for now I think it makes sense to
# expose that functionality as a function for easier later deprecation.
def cqm_to_bqm(cqm: ConstrainedQuadraticModel, lagrange_multiplier: Optional[Bias] = None,
               ) -> Tuple[BinaryQuadraticModel, CQMToBQMInverter]:
    """Construct a binary quadratic model from a constrained quadratic model.

    Args:
        cqm: A constrained quadratic model. All constraints must be linear
            and all integer variables must have a lower bound of 0.

        lagrange_multiplier: The penalty strength used when converting
            constraints into penalty models. Defaults to 10x the largest
            bias in the objective.

    Returns:
        A 2-tuple containing:

            A binary quadratic model

            A function that converts samples over the binary quadratic model
            back into samples for the constrained quadratic model.

    Example:

        Start with a constrained quadratic model

        >>> num_widget_a = dimod.Integer('num_widget_a', upper_bound=7)
        >>> num_widget_b = dimod.Integer('num_widget_b', upper_bound=3)
        >>> cqm = dimod.ConstrainedQuadraticModel()
        >>> cqm.set_objective(-3 * num_widget_a - 4 * num_widget_b)
        >>> cqm.add_constraint(num_widget_a + num_widget_b <= 5, label='total widgets')
        'total widgets'

        Convert it to a binary quadratic model and solve it using
        :class:`dimod.ExactSolver`.

        >>> bqm, invert = dimod.cqm_to_bqm(cqm)
        >>> sampleset = dimod.ExactSolver().sample(bqm)

        Interpret the answer in the original variable classes

        >>> invert(sampleset.first.sample)
        {'num_widget_a': 2, 'num_widget_b': 3}

        Note that the inverter is also serializable.

        >>> import json
        >>> newinvert = dimod.constrained.CQMToBQMInverter.from_dict(
        ...     json.loads(json.dumps(invert.to_dict())))
        >>> newinvert(sampleset.first.sample)
        {'num_widget_a': 2, 'num_widget_b': 3}

    """

    from dimod.generators.integer import binary_encoding  # avoid circular import

    bqm = BinaryQuadraticModel(Vartype.BINARY)
    binary: Dict[Variable, Vartype] = {}
    integers: Dict[Variable, BinaryQuadraticModel] = {}

    # add the variables
    for v in cqm.variables:
        vartype = cqm.vartype(v)

        if vartype is Vartype.SPIN or vartype is Vartype.BINARY:
            binary[v] = vartype
        elif vartype is Vartype.INTEGER:
            if cqm.lower_bound(v) != 0:
                raise ValueError("integer variables must have a lower bound of 0, "
                                 f"variable {v} has a lower bound of {cqm.lower_bound(v)}")
            v_bqm = integers[v] = binary_encoding(v, int(cqm.upper_bound(v)))

            if not v_bqm.variables.isdisjoint(bqm.variables):
                # this should be pretty unusual, so let's not bend over backwards
                # to accommodate it.
                raise ValueError("given CQM has conflicting variables with ones "
                                 "generated by dimod.generators.binary_encoding")

            bqm.add_variables_from((v, 0) for v in v_bqm.variables)
        else:
            raise RuntimeError("unexpected vartype")

    # objective, we know it's always a QM
    bqm += _qm_to_bqm(cqm.objective, integers)

    if lagrange_multiplier is None:
        if cqm.constraints and bqm.num_variables:
            max_bias = max(-bqm.linear.min(), bqm.linear.max())
            if not bqm.is_linear():
                max_bias = max(-bqm.quadratic.min(), bqm.quadratic.max(), max_bias)
            lagrange_multiplier = 10 * max_bias
        else:
            lagrange_multiplier = 0  # doesn't matter

    for constraint in cqm.constraints.values():
        lhs = constraint.lhs
        rhs = constraint.rhs
        sense = constraint.sense

        if isinstance(lhs, QuadraticModel):
            lhs = _qm_to_bqm(lhs, integers)

        if not lhs.is_linear():
            raise ValueError("CQM must not have any quadratic constraints")

        if lhs.vartype is Vartype.SPIN:
            lhs = lhs.change_vartype(Vartype.BINARY, inplace=True)

        # at this point we know we have a BINARY bqm

        if sense is Sense.Eq:
            bqm.add_linear_equality_constraint(
                ((v, lhs.get_linear(v)) for v in lhs.variables),
                lagrange_multiplier,
                lhs.offset - rhs,
                )
        elif sense is Sense.Ge:
            bqm.add_linear_inequality_constraint(
                ((v, lhs.get_linear(v)) for v in lhs.variables),
                lagrange_multiplier,
                new_label(),
                constant=lhs.offset,
                lb=rhs,
                ub=np.iinfo(np.int64).max,
                )
        elif sense is Sense.Le:
            bqm.add_linear_inequality_constraint(
                ((v, lhs.get_linear(v)) for v in lhs.variables),
                lagrange_multiplier,
                new_label(),
                constant=lhs.offset,
                lb=np.iinfo(np.int64).min,
                ub=rhs,
                )
        else:
            raise RuntimeError("unexpected sense")

    return bqm, CQMToBQMInverter(binary, integers)


# register fileview loader
load.register(CQM_MAGIC_PREFIX, ConstrainedQuadraticModel.from_file)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections.abc as abc
import warnings

from collections import OrderedDict
from copy import deepcopy
from numbers import Integral

import numpy as np

from dimod.core.bqm import BQM, ShapeableBQM
from dimod.utilities import iter_safe_relabels
from dimod.vartypes import as_vartype, Vartype

try:
    from dimod.bqm.common import itype, ntype
except ImportError:
    itype = np.dtype(np.uint32)
    ntype = np.dtype(np.uint64)


__all__ = ['AdjDictBQM']


class AdjDictBQM(ShapeableBQM):
    """A binary quadratic model structured as a dict-of-dicts.

    `AdjVectorBQM` is deprecated. Use :class:`BinaryQuadraticModel` with
    `object` dtype instead.

    Can be created in several ways:

        AdjDictBQM(vartype)
            Creates an empty binary quadratic model (BQM).

        AdjDictBQM(bqm)
            Creates a BQM from another BQM. See `copy` and `cls` kwargs below.

        AdjDictBQM(bqm, vartype)
            Creates a BQM from another BQM, changing to the appropriate
            `vartype` if necessary.

        AdjDictBQM(n, vartype)
            Creates a BQM with `n` variables, indexed linearly from zero,
            setting all biases to zero.

        AdjDictBQM(quadratic, vartype)
            Creates a BQM from quadratic biases given as a square array_like_
            or a dictionary of the form `{(u, v): b, ...}`. Note that when
            formed with SPIN-variables, biases on the diagonal are added to the
            offset.

        AdjDictBQM(linear, quadratic, vartype)
            Creates a BQM from linear and quadratic biases, where `linear` is a
            one-dimensional array_like_ or a dictionary of the form
            `{v: b, ...}`, and `quadratic` is a square array_like_ or a
            dictionary of the form `{(u, v): b, ...}`. Note that when formed
            with SPIN-variables, biases on the diagonal are added to the offset.

        AdjDictBQM(linear, quadratic, offset, vartype)
            Creates a BQM from linear and quadratic biases, where `linear` is a
            one-dimensional array_like_ or a dictionary of the form
            `{v: b, ...}`, and `quadratic` is a square array_like_ or a
            dictionary of the form `{(u, v): b, ...}`, and `offset` is a
            numerical offset. Note that when formed with SPIN-variables, biases
            on the diagonal are added to the offset.

    Notes:

        The AdjDictBQM is implemented using a dict-of-dicts structure. The
        outer dict contains the BQM's variables as keys and the neighborhoods
        as values. Each neighborhood dict contains the neighbors as keys and
        the quadratic biases as values. The linear biases are stored as
        self-interactions.

        Advantages:

        - Pure python implementation
        - Supports arbitrary python types as biases
        - Low complexity for lookup operations
        - Supports incremental construction

        Disadvantages:

        - Slow iteration
        - High memory usage

        Intended Use:

        - For small problems or when flexibility is important

    Examples:
        The first example constructs a BQM from a dict.

        >>> dimod.AdjDictBQM({'a': -1.0}, {('a', 'b'): 1.0}, 'SPIN')
        AdjDictBQM({a: -1.0, b: 0.0}, {('a', 'b'): 1.0}, 0.0, 'SPIN')

        The next example demonstrates incremental construction:

        >>> bqm = dimod.AdjDictBQM('SPIN')
        >>> bqm.add_variable('a')
        'a'
        >>> bqm.add_variable()
        1
        >>> bqm.set_quadratic('a', 1, 3.0)
        >>> bqm
        AdjDictBQM({a: 0.0, 1: 0.0}, {('a', 1): 3.0}, 0.0, 'SPIN')

        This example shows support for arbitrary types.

        >>> import numpy as np
        >>> from fractions import Fraction
        >>> dimod.AdjDictBQM({('a', 'b'): Fraction(1, 3)}, 'BINARY')
        AdjDictBQM({a: 0.0, b: 0.0}, {('a', 'b'): 1/3}, 0.0, 'BINARY')

    .. _array_like: https://numpy.org/doc/stable/user/basics.creation.html

    """
    # developer note: we should add a FAQ for why spin-valued diagonal biases
    # are offsets and reference it here

    def __init__(self, *args, vartype=None):
        warnings.warn("AdjDictBQM(..) is deprecated and will be removed in dimod 0.11.0, "
                      "please use BinaryQuadratic(.., dtype=object) instead",
                      DeprecationWarning, stacklevel=2)

        if vartype is not None:
            # pass in as a positional argument
            self.__init__(*args, vartype)
            return

        self.dtype = np.dtype(object)

        # we use ordered dict because the other BQM types are ordered. However
        # collection's OrderedDict is optimized for both FILO and FIFO so has
        # a larger memory footprint than we need. We could instead store a list
        # with the dict which would be more performant but complicate the
        # implementation
        self._adj = adj = OrderedDict()

        if len(args) == 0:
            raise TypeError("A valid vartype or another bqm must be provided")
        if len(args) == 1:
            # BQM(bqm) or BQM(vartype)
            obj, = args
            if isinstance(obj, BQM):
                self._init_bqm(obj)
            else:
                self._init_number(0, obj)
        elif len(args) == 2:
            # BQM(bqm, vartype), BQM(n, vartype) or BQM(M, vartype)
            obj, vartype = args
            if isinstance(obj, BQM):
                self._init_bqm(obj, vartype)
            elif isinstance(obj, Integral):
                self._init_number(obj, vartype)
            else:
                self._init_components({}, obj, 0.0, vartype)
        elif len(args) == 3:
            # BQM(linear, quadratic, vartype)
            linear, quadratic, vartype = args
            self._init_components(linear, quadratic, 0.0, vartype)
        elif len(args) == 4:
            # BQM(linear, quadratic, offset, vartype)
            self._init_components(*args)
        else:
            msg = "__init__() takes 4 positional arguments but {} were given"
            raise TypeError(msg.format(len(args)))

    def _init_bqm(self, bqm, vartype=None):
        self.linear.update(bqm.linear)
        self.quadratic.update(bqm.quadratic)
        self.offset = bqm.offset
        self._vartype = bqm.vartype

        if vartype is not None:
            self.change_vartype(as_vartype(vartype), inplace=True)

    def _init_components(self, linear, quadratic, offset, vartype):
        self._vartype = vartype = as_vartype(vartype)

        if isinstance(linear, (abc.Mapping, abc.Iterator)):
            self.linear.update(linear)
        else:
            # assume a sequence
            self.linear.update(enumerate(linear))

        adj = self._adj

        if isinstance(quadratic, abc.Mapping):
            for (u, v), bias in quadratic.items():
                self.add_variable(u)
                self.add_variable(v)

                if u == v and vartype is Vartype.SPIN:
                    offset = offset + bias  # not += on off-chance it's mutable
                elif u in adj[v]:
                    adj[u][v] = adj[v][u] = adj[u][v] + bias
                else:
                    adj[u][v] = adj[v][u] = bias
        elif isinstance(quadratic, abc.Iterator):
            for u, v, bias in quadratic:
                self.add_variable(u)
                self.add_variable(v)

                if u == v and vartype is Vartype.SPIN:
                    offset = offset + bias  # not += on off-chance it's mutable
                elif u in adj[v]:
                    adj[u][v] = adj[v][u] = adj[u][v] + bias
                else:
                    adj[u][v] = adj[v][u] = bias
        else:
            # unlike the other BQM types we let numpy handle the typing
            if isinstance(quadratic, np.ndarray):
                dtype = quadratic.dtype
            else:
                quadratic = np.asarray(quadratic, dtype=object)

            D = np.atleast_2d(quadratic)

            num_variables = D.shape[0]

            if D.ndim != 2 or num_variables != D.shape[1]:
                raise ValueError("expected dense to be a 2 dim square array")

            # make sure all the variables are present
            for v in range(num_variables):
                self.add_variable(v)

            it = np.nditer(D, flags=['multi_index', 'refs_ok'], op_flags=['readonly'])
            while not it.finished:
                u, v = it.multi_index
                bias = it.value[()]

                if bias:
                    if u == v and vartype is Vartype.SPIN:
                        # not += on off-chance it's mutable
                        offset = offset + bias
                    elif u in adj[v]:
                        adj[u][v] = adj[v][u] = adj[u][v] + bias
                    else:
                        adj[u][v] = adj[v][u] = bias

                it.iternext()

        self.offset = offset

    def _init_number(self, n, vartype):
        self.linear.update((v, 0.0) for v in range(n))
        self.offset = 0.0
        self._vartype = as_vartype(vartype)

    @property
    def num_variables(self):
        """int: The number of variables in the model."""
        return len(self._adj)

    @property
    def num_interactions(self):
        """int: The number of interactions in the model."""
        return (sum(map(len, self._adj.values())) - len(self._adj)) // 2

    @property
    def offset(self):
        """The constant energy offset associated with the model."""
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset

    @property
    def vartype(self):
        """:class:`.Vartype`: The vartype of the binary quadratic model. One of
        :class:`.Vartype.SPIN` or :class:`.Vartype.BINARY`.
        """
        # so it's readonly
        return self._vartype

    def add_variable(self, v=None, bias=0.0):
        """Add a variable to the binary quadratic model.

        Args:
            v (hashable, optional):
                A label for the variable. Defaults to the length of the binary
                quadratic model, if that label is available. Otherwise defaults
                to the lowest available positive integer label.

            bias (numeric, optional, default=0):
                The initial bias value for the added variable. If `v` is already
                a variable, then `bias` (if any) is adding to its existing
                linear bias.

        Returns:
            hashable: The label of the added variable.

        Raises:
            TypeError: If the label is not hashable.

        Examples:

            >>> bqm = dimod.AdjDictBQM('SPIN')
            >>> bqm.add_variable()
            0
            >>> bqm.add_variable('a')
            'a'
            >>> bqm.add_variable()
            2

            >>> bqm = dimod.AdjDictBQM('SPIN')
            >>> bqm.add_variable(1)
            1
            >>> bqm.add_variable()  # 1 is taken
            0
            >>> bqm.add_variable()
            2

        """
        if v is None:

            v = self.num_variables

            if self.has_variable(v):
                # if v already exists there must be a smaller one that's free
                for v in range(self.num_variables):
                    if not self.has_variable(v):
                        break

        self._adj.setdefault(v, OrderedDict({v: 0.0}))[v] += bias
        return v

    def change_vartype(self, vartype, inplace=True):
        """Return a binary quadratic model with the specified vartype.

        Args:
            vartype (:class:`.Vartype`/str/set, optional):
                Variable type for the changed model. Accepted input values:

                * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

            inplace (bool, optional, default=True):
                If True, the binary quadratic model is updated in-place;
                otherwise, a new binary quadratic model is returned.

        Returns:
            :obj:`.AdjDictBQM`: A binary quadratic model with the specified
            vartype.

        """
        if not inplace:
            return self.copy().change_vartype(vartype, inplace=True)

        vartype = as_vartype(vartype)

        # in place and we are already correct, so nothing to do
        if self.vartype == vartype:
            return self

        if vartype is Vartype.BINARY:
            lin_mp, lin_offset_mp = 2.0, -1.0
            quad_mp, lin_quad_mp, quad_offset_mp = 4.0, -2.0, 1.0
        elif vartype is Vartype.SPIN:
            lin_mp, lin_offset_mp = 0.5, .5
            quad_mp, lin_quad_mp, quad_offset_mp = 0.25, 0.25, 0.25
        else:
            raise ValueError("unkown vartype")

        for v, bias in self.linear.items():
            self.linear[v] = lin_mp * bias
            self.offset += lin_offset_mp * bias

        for (u, v), bias in self.quadratic.items():
            self.adj[u][v] = quad_mp * bias

            self.linear[u] += lin_quad_mp * bias
            self.linear[v] += lin_quad_mp * bias

            self.offset += quad_offset_mp * bias

        self._vartype = vartype

        return self

    def __copy__(self):
        # deep copy is more than we need but it works and this class is
        # deprecated anyway
        return deepcopy(self)

    def degree(self, v):
        """Return degree of the specified variable.

        The degree is the number of variables sharing an interaction with ``v``.

        Args:
            v (hashable):
                Variable in the binary quadratic model.

        Returns:
            Degree of ``v``.

        Raises:
            ValueError: If ``v`` is not a variable in the binary quadratic model.
        """
        try:
            return len(self._adj[v]) - 1
        except KeyError:
            raise ValueError("{} is not a variable".format(v))

    def get_linear(self, v):
        """Get the linear bias of the specified variable.

        Args:
            v (hashable):
                A variable in the binary quadratic model.

        Returns:
            float: The linear bias of ``v``.

        Raises:
            ValueError: If ``v`` is not a variable in the binary quadratic model.

        """
        try:
            return self._adj[v][v]
        except KeyError:
            pass
        raise ValueError("{} is not a variable".format(v))

    def get_quadratic(self, u, v, default=None):
        """Get the quadratic bias of the specified interaction.

        Args:
            u (hashable):
                A variable in the binary quadratic model.

            v (hashable):
                A variable in the binary quadratic model.

            default (optional):
                Value to return if there is no interactions between ``u`` and ``v``.

        Returns:
            The quadratic bias of ``(u, v)``.

        Raises:
            ValueError: If either ``u`` or ``v`` is not a variable in the binary
                quadratic model or if ``u == v``

            ValueError: If ``(u, v)`` is not an interaction and `default` is
                `None`.

        """
        if u == v:
            raise ValueError("No interaction between {} and itself".format(u))
        try:
            return self._adj[u][v]
        except KeyError:
            pass
        if default is not None:
            return default
        raise ValueError('No interaction between {} and {}'.format(u, v))

    def iter_linear(self):
        """Iterate over the linear biases of the binary quadratic model.

        Yields:
            tuple: A variable in the binary quadratic model and its linear bias.
        """
        for u, neighborhood in self._adj.items():
            yield u, neighborhood[u]  # u and it's bias

    def iter_quadratic(self, variables=None):
        """Iterate over the quadratic biases of the binary quadratic model.

        Args:
            variables (iterable):
                Variables in the binary quadratic model. Iterates only over
                interactions of these variables.

        Yields:
            3-tuple: Interaction variables in the binary quadratic model and their
            bias.
        """

        if variables is None:
            variables = self._adj
        elif self.has_variable(variables):
            variables = [variables]

        seen = set()
        for u in variables:
            neighborhood = self._adj[u]
            seen.add(u)  # also avoids self-loops
            for v, bias in neighborhood.items():
                if v not in seen:
                    yield (u, v, bias)

    def remove_variable(self, v=None):
        """Remove a variable and its associated interactions.

        Args:
            v (variable, optional):
                The variable to be removed from the binary quadratic model
                (BQM). If not provided, the last variable added is removed.

        Returns:
            variable: The removed variable.

        Raises:
            ValueError: If the BQM is empty or if ``v`` is not
                a variable.

        """
        if len(self._adj) == 0:
            raise ValueError("pop from empty binary quadratic model")

        if v is None:
            v, neighborhood = self._adj.popitem(last=True)
        else:
            try:
                neighborhood = self._adj.pop(v)
            except KeyError:
                raise ValueError('{!r} is not a variable'.format(v))

        for u in neighborhood:
            if u != v:
                self._adj[u].pop(v)

        return v

    def remove_interaction(self, u, v):
        """Remove the interaction between the specified variables.

        Args:
            u (hashable):
                A variable in the binary quadratic model.

            v (hashable):
                A variable in the binary quadratic model.

        Returns:
            bool: If there was an interaction to remove.

        Raises:
            ValueError: If either ``u`` or ``v`` is not a variable in the binary
                quadratic model.

        """
        if u == v:
            raise ValueError("No interaction between {} and itself".format(u))

        try:
            self._adj[u].pop(v)
        except KeyError:
            # nothing pop
            raise ValueError('No interaction between {} and {}'.format(u, v))

        # want to raise an exception in the case that they got out of sync
        # as a sanity check
        self._adj[v].pop(u)

        return True

    def relabel_variables(self, mapping, inplace=True):
        """Relabel variables of a binary quadratic model as specified by mapping.

        Args:
            mapping (dict):
                Dict mapping current variable labels to new ones. If an
                incomplete mapping is provided, unmapped variables retain their
                current labels.

            inplace (bool, optional, default=True):
                If True, the binary quadratic model is updated in-place;
                otherwise, a new binary quadratic model is returned.

        Returns:
            A binary quadratic model with the variables relabeled. If `inplace`
            is set to True, returns itself.

        """
        if not inplace:
            return self.copy().relabel_variables(mapping, inplace=True)

        adj = self._adj

        for submap in iter_safe_relabels(mapping, self.linear):
            for old, new in submap.items():
                if old == new:
                    continue

                # replace the linear bias
                adj[new] = {new: adj[old].pop(old)}

                # copy the quadratic biases
                for v in adj[old]:
                    adj[new][v] = adj[v][new] = adj[v].pop(old)

                # remove the old adj for old
                del adj[old]

        return self

    def set_linear(self, v, bias):
        """Set the linear biase of the specified variable.

        Args:
            v (hashable):
                A variable in the binary quadratic model. It is added if not
                already in the model.

            b (numeric):
                The linear bias of ``v``.

        Raises:
            TypeError: If ``v`` is not hashable

        """
        if v in self._adj:
            self._adj[v][v] = bias
        else:
            self._adj[v] = {v: bias}

    def set_quadratic(self, u, v, bias):
        """Set the quadratic bias of an interaction specified by its variables.

        Args:
            u (hashable):
                A variable in the binary quadratic model.

            v (hashable):
                A variable in the binary quadratic model.

            b (numeric):
                The quadratic bias to set for interaction ``(u, v)``

        Raises:
            TypeError: If ``u`` or ``v`` is not hashable.

        """
        # make sure the variables exist
        self.add_variable(u)
        self.add_variable(v)

        adj = self._adj
        adj[u][v] = adj[v][u] = bias



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""

A structured sampler can only sample from binary quadratic models with a specific graph.

For structured samplers you must implement the :attr:`~.Structured.nodelist` and
:attr:`~.Structured.edgelist` properties. The :class:`.Structured` abstract base
class provides access to the :attr:`~.Structured.structure` and
:attr:`~.Structured.adjacency` properties, the
:meth:`~.Structured.to_networkx_graph` method, as well as any method or
properties required by the :class:`.Sampler` abstract base class. The
:obj:`~.bqm_structured` decorator verifies that any given binary quadratic model
conforms to the supported structure.

Examples:
    This simple example shows a structured sampler that can only sample from a binary
    quadratic model with two variables and one interaction.

    .. testcode::

        class TwoVariablesSampler(dimod.Sampler, dimod.Structured):
            @property
            def nodelist(self):
                return [0, 1]

            @property
            def edgelist(self):
                return [(0, 1)]

            @property
            def properties(self):
                return dict()

            @property
            def parameters(self):
                return dict()

            @dimod.decorators.bqm_structured
            def sample(self, bqm):
                # All bqm's passed in will be a subgraph of the sampler's structure
                variable_list = list(bqm.linear)
                samples = []
                energies = []
                for values in itertools.product(bqm.vartype.value, repeat=len(bqm)):
                    sample = dict(zip(variable_list, values))
                    samples.append(sample)
                    energies.append(bqm.energy(sample))

                return dimod.SampleSet.from_samples(samples, bqm.vartype, energies)

                return response

    >>> import itertools
    >>> sampler = TwoVariablesSampler()
    >>> solutions = sampler.sample_ising({}, {(0, 1): -1})
    >>> solutions.first.energy
    -1.0

"""
from __future__ import annotations

from collections import namedtuple

import abc
import typing

from dimod.typing import Variable

if typing.TYPE_CHECKING:
    from dimod import BinaryQuadraticModel

__all__ = ['Structured']

class _Structure(typing.NamedTuple):
    nodelist: typing.List[Variable]
    edgelist: typing.List[typing.Tuple[Variable, Variable]]
    adjacency: typing.Dict[Variable, typing.Set]

class Structured(abc.ABC):
    """The abstract base class for dimod structured samplers.

    Provides the :attr:`.Structured.adjacency` and :attr:`.Structured.structure`
    properties, and the :meth:`.Structured.to_networkx_graph` method.

    Abstract properties :attr:`~.Structured.nodelist` and :attr:`~.Structured.edgelist`
    must be implemented.

    """
    @abc.abstractproperty
    def nodelist(self) -> typing.List[Variable]:
        """Nodes/variables allowed by the sampler."""
        pass

    @abc.abstractproperty
    def edgelist(self) -> typing.List[typing.Tuple[Variable, Variable]]:
        """Edges/interactions allowed by the sampler.
        """
        pass

    @property
    def adjacency(self) -> typing.Dict[Variable, typing.Set]:
        """Adjacency structure formatted as a dict, where keys are the nodes of
        the structured sampler and values are sets of all adjacent nodes for each
        key node.
        """
        if not hasattr(self, '_adjacency'):
            adjacency = {v: set() for v in self.nodelist}
            for u, v in self.edgelist:
                if v in adjacency[u]:
                    raise ValueError("Each edge in edgelist must be unique")
                adjacency[u].add(v)
                adjacency[v].add(u)
            self._adjacency = adjacency
            return adjacency
        return self._adjacency

    @property
    def structure(self) -> _Structure:
        """Structure of the structured sampler formatted as a
        :func:`~collections.namedtuple` where the 3-tuple values are the
        :attr:`.nodelist`, :attr:`.edgelist` and :attr:`.adjacency` attributes.
        """
        return _Structure(self.nodelist, self.edgelist, self.adjacency)

    def to_networkx_graph(self):
        """Convert structure to NetworkX graph format.

        Note that NetworkX must be installed for this method to work.

        Returns:
            :class:`networkx.Graph`: A NetworkX graph containing the nodes and
            edges from the sampler's structure.

        """
        import networkx as nx

        G = nx.Graph(self.edgelist)
        G.add_nodes_from(self.nodelist)

        return G

    def valid_bqm_graph(self, bqm: BinaryQuadraticModel) -> bool:
        """Validate that problem defined by :class:`dimod.BinaryQuadraticModel`
        matches the graph provided by the sampler.

        Args:
            bqm: :class:`dimod.BinaryQuadraticModel` object to validate.

        Returns:
            Boolean indicating validity of BQM graph compared to sampler graph.

        """
        return all(variable in self.adjacency for variable in bqm.variables) and \
               all(neighbor in self.adjacency[vertex] for neighbor, vertex in bqm.quadratic)



# distutils: language = c++
# cython: language_level=3
#
# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from dimod.bqm.cppbqm cimport AdjVectorBQM as cppAdjVectorBQM
from dimod.bqm.common cimport VarIndex, Bias

cdef class cyAdjVectorBQM:
    """A binary quadratic model where the neighborhoods are C++ vectors.

    `AdjVectorBQM` is deprecated. Use :class:`BinaryQuadraticModel` instead.

    Can be created in several ways:

        AdjVectorBQM(vartype)
            Creates an empty binary quadratic model (BQM).

        AdjVectorBQM(bqm)
            Creates a BQM from another BQM. See `copy` and `cls` kwargs below.

        AdjVectorBQM(bqm, vartype)
            Creates a BQM from another BQM, changing to the appropriate
            `vartype` if necessary.

        AdjVectorBQM(n, vartype)
            Creates a BQM with `n` variables, indexed linearly from zero,
            setting all biases to zero.

        AdjVectorBQM(quadratic, vartype)
            Creates a BQM from quadratic biases given as a square array_like_
            or a dictionary of the form `{(u, v): b, ...}`. Note that when
            formed with SPIN-variables, biases on the diagonal are added to the
            offset.

        AdjVectorBQM(linear, quadratic, vartype)
            Creates a BQM from linear and quadratic biases, where `linear` is a
            one-dimensional array_like_ or a dictionary of the form
            `{v: b, ...}`, and `quadratic` is a square array_like_ or a
            dictionary of the form `{(u, v): b, ...}`. Note that when formed
            with SPIN-variables, biases on the diagonal are added to the offset.

        AdjVectorBQM(linear, quadratic, offset, vartype)
            Creates a BQM from linear and quadratic biases, where `linear` is a
            one-dimensional array_like_ or a dictionary of the form
            `{v: b, ...}`, and `quadratic` is a square array_like_ or a
            dictionary of the form `{(u, v): b, ...}`, and `offset` is a
            numerical offset. Note that when formed with SPIN-variables, biases
            on the diagonal are added to the offset.

    Notes:

        The AdjVectorBQM is implemented using an adjacency structure where the
        neighborhoods are implemented as C++ vectors.

        Advantages:

        - Supports incremental construction
        - Fast iteration over the biases

        Disadvantages:

        - Only supports float64 biases

        Intended Use:

        - When performance is important and the use case requires incremental
          construction
        - This should be the default BQM type for large problems where
          arbitrary types are not needed

    Examples:

        >>> import numpy as np
        >>> from dimod import AdjVectorBQM

        Construct from dicts.

        >>> AdjVectorBQM({'a': -1.0}, {('a', 'b'): 1.0}, 'SPIN')
        AdjVectorBQM({a: -1.0, b: 0.0}, {('a', 'b'): 1.0}, 0.0, 'SPIN')

        Incremental Construction.

        >>> bqm = AdjVectorBQM('SPIN')
        >>> bqm.add_variable('a')
        'a'
        >>> bqm.add_variable()
        1
        >>> bqm.set_quadratic('a', 1, 3.0)
        >>> bqm
        AdjVectorBQM({a: 0.0, 1: 0.0}, {('a', 1): 3.0}, 0.0, 'SPIN')

    .. _array_like: https://numpy.org/doc/stable/user/basics.creation.html

    """
    cdef cppAdjVectorBQM[VarIndex, Bias] bqm_

    cdef Bias offset_

    cdef readonly object vartype
    """Variable type, :class:`.Vartype.SPIN` or :class:`.Vartype.BINARY`."""

    cdef readonly object dtype
    """Data type of the linear biases, float64."""

    cdef readonly object itype
    """Data type of the indices, uint32."""

    cdef readonly object ntype
    """Data type of the neighborhood indices, varies by platform."""

    # these are not public because the user has no way to access the underlying
    # variable indices
    cdef object _label_to_idx
    cdef object _idx_to_label

    cdef VarIndex label_to_idx(self, object) except *



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import typing

from collections.abc import Collection, KeysView, Callable
from functools import reduce
from typing import Any, Dict, Iterator, Optional, Tuple, Mapping

import numpy as np

try:
    from numpy.typing import ArrayLike, DTypeLike
except ImportError:
    ArrayLike = Any
    DTypeLike = Any

from dimod.sampleset import as_samples
from dimod.typing import Variable, VartypeLike
from dimod.utilities import iter_safe_relabels
from dimod.vartypes import as_vartype, Vartype


class pyBQM:
    """A pure-python BQM implementation for handling arbitrary bias types."""
    def __init__(self, vartype: VartypeLike):
        self._adj: Dict[Variable, Dict[Variable, Any]] = dict()
        self._vartype = as_vartype(vartype)

        self.offset = 0

    @property
    def dtype(self) -> np.dtype:
        return np.dtype('O')

    @property
    def variables(self) -> Collection:
        return KeysView(self._adj)

    def __copy__(self):
        new = type(self)(self._vartype)
        adj = new._adj
        for v, neighborhood in self._adj.items():
            adj[v] = neighborhood.copy()
        new.offset = self.offset
        return new

    def add_linear(self, v: Variable, bias: Any):
        self._adj.setdefault(v, dict())
        try:
            zero = type(bias)()  # try to preserve the type
        except TypeError:
            zero = 0  # sometimes it cannot be constructed with no arguments
        self._adj[v][v] = self._adj[v].get(v, zero) + bias

    def add_linear_equality_constraint(self, *args, **kwargs):
        raise NotImplementedError  # defer to caller

    def add_linear_from_array(self, linear: ArrayLike):
        for v, bias in enumerate(np.asarray(linear)):
            self.add_linear(v, bias)

    def add_quadratic(self, u: Variable, v: Variable, bias: Any):
        if u == v:
            raise ValueError(f"{u!r} cannot have an interaction with itself")

        try:
            zero = type(bias)()  # try to preserve the type
        except TypeError:
            zero = 0  # sometimes it cannot be constructed with no arguments

        # derive the linear types to match quadratic. This might not always
        # be what we want but it's as good a guess as any
        if u not in self.variables:
            self.set_linear(u, zero)
        if v not in self.variables:
            self.set_linear(v, zero)

        self._adj[u][v] = self._adj[v][u] = self._adj[v].get(u, zero) + bias

    def add_quadratic_from_dense(self, quadratic: ArrayLike):
        quadratic = np.asarray(quadratic)
        if quadratic.shape[0] != quadratic.shape[1]:
            raise ValueError("quadratic must be a square matrix")

        num_variables = quadratic.shape[0]

        for u in range(num_variables):
            for v in range(num_variables):
                if u == v:
                    continue
                elif quadratic[u, v]:
                    self.add_quadratic(u, v, quadratic[u, v])

        # now handle the linear
        if self._vartype is Vartype.SPIN:
            for v in range(num_variables):
                # since s*s == 1
                self.offset += quadratic[v, v]
        elif self._vartype is Vartype.BINARY:
            for v in range(num_variables):
                # since x*x == x
                self.add_linear(v, quadratic[v, v])
        else:
            raise RuntimeError("unexpected vartype")

    def add_variable(self, v: Optional[Variable] = None,
                     bias: Any = 0) -> Variable:
        if v is None:
            # match the behaviour of dimod.Variables
            adj = self._adj

            v = len(adj)

            if v in adj:
                v = 0
                while v in adj:
                    v += 1

        self.add_linear(v, bias)
        return v

    def change_vartype(self, vartype: VartypeLike) -> 'pyBQM':
        vartype = as_vartype(vartype)

        # in place and we are already correct, so nothing to do
        if self._vartype == vartype:
            return self

        if vartype == Vartype.BINARY:
            lin_mp = 2.
            lin_offset_mp = -1.
            quad_mp = 4.
            lin_quad_mp = -2.
            quad_offset_mp = .5
        elif vartype == Vartype.SPIN:
            lin_mp = .5
            lin_offset_mp = .5
            quad_mp = .25
            lin_quad_mp = .25
            quad_offset_mp = .125
        else:
            raise RuntimeError("unexpected vartype")

        adj = self._adj

        for u, Nu in adj.items():
            lbias = Nu[u]

            self.offset += lin_offset_mp * lbias
            Nu[u] = lin_mp * lbias

            for v, qbias in Nu.items():
                if v == u:
                    continue
                Nu[v] = quad_mp * qbias
                Nu[u] += lin_quad_mp * qbias  # linear
                self.offset += quad_offset_mp * qbias

        self._vartype = vartype

        return self

    def degree(self, v: Variable) -> int:
        try:
            return len(self._adj[v]) - 1
        except KeyError:
            raise ValueError(f"unknown variable {v!r}") from None

    def energies(self, samples_like, dtype: DTypeLike = None):
        samples, labels = as_samples(samples_like)

        bqm_to_sample = dict((v, i) for i, v in enumerate(labels))

        if not bqm_to_sample.keys() >= self._adj.keys():
            raise ValueError(
                f"missing variable {(self._adj.keys() - bqm_to_sample.keys()).pop()!r} in sample(s)"
                )

        ldata = np.asarray([self.get_linear(v) if v in self._adj else 0 for v in labels],
                           dtype=dtype)

        # let's use floats (or the more general object) if dtype is not specified
        # so that we do proper type promotion
        if dtype is None and np.issubdtype(ldata.dtype, np.number):
            ldata = np.asarray(ldata, dtype=np.float64)

        irow = []
        icol = []
        qdata = []
        for u, v, bias in self.iter_quadratic():
            irow.append(bqm_to_sample[u])
            icol.append(bqm_to_sample[v])
            qdata.append(bias)

        energies = samples.dot(ldata)
        energies += (samples[:, irow]*samples[:, icol]).dot(qdata)
        energies += energies.dtype.type(self.offset)

        return np.asarray(energies, dtype=dtype)

    def get_linear(self, v: Variable) -> Any:
        try:
            return self._adj[v][v]
        except KeyError:
            raise ValueError(f"unknown variable {v!r}") from None

    def get_quadratic(self, u: Variable, v: Variable,
                      default: Optional[Any] = None) -> Any:
        if u == v:
            raise ValueError(f"{u!r} cannot have an interaction with itself")

        try:
            return self._adj[u][v]
        except KeyError:
            if default is None:
                raise ValueError(
                    f"{u!r} and {v!r} have no interaction") from None
            return default

    def is_linear(self) -> bool:
        return not any(len(Nv) > 1 for Nv in self._adj.values())

    def iter_neighborhood(self, v: Variable) -> Iterator[Any]:
        try:
            Nv = self._adj[v]
        except KeyError:
            raise ValueError(f"unknown variable {v!r}") from None
        for u, bias in Nv.items():
            if u != v:
                yield u, bias

    def iter_quadratic(self) -> Iterator[Tuple[Variable, Variable, Any]]:
        seen = set()
        for u, Nu in self._adj.items():
            seen.add(u)
            for v, bias in Nu.items():
                if v not in seen:
                    yield u, v, bias

    def nbytes(self, *args, **kwargs) -> typing.NoReturn:
        raise TypeError(
            "cannot return the number of bytes for a binary quadratic model with object dtype")

    def num_variables(self) -> int:
        return len(self._adj)

    def num_interactions(self) -> int:
        n = sum(map(len, self._adj.values()), 0)
        n -= self.num_variables()  # subtract the self-loops
        return n // 2

    def reduce_linear(self, function: Callable,
                      initializer: Optional[Any] = None) -> Any:
        gen = (self.get_linear(v) for v in self.variables)
        if initializer is None:
            return reduce(function, gen)
        else:
            return reduce(function, gen, initializer)

    def reduce_neighborhood(self, v: Variable, function: Callable,
                            initializer: Optional[Any] = None) -> Any:
        gen = (b for _, b in self.iter_neighborhood(v))
        if initializer is None:
            return reduce(function, gen)
        else:
            return reduce(function, gen, initializer)

    def reduce_quadratic(self, function: Callable,
                         initializer: Optional[Any] = None) -> Any:
        gen = (b for _, _, b in self.iter_quadratic())
        if initializer is None:
            return reduce(function, gen)
        else:
            return reduce(function, gen, initializer)

    def remove_interaction(self, u: Variable, v: Variable):
        if u == v:
            raise ValueError(f"{u!r} cannot have an interaction with itself")
        try:
            self._adj[u].pop(v)
        except KeyError:
            raise ValueError(f"{u!r} and {v!r} have no interaction") from None

        self._adj[v].pop(u)

    def relabel_variables(self, mapping: Mapping[Variable, Variable]):
        adj = self._adj

        for submap in iter_safe_relabels(mapping, self.variables):
            for old, new in submap.items():
                if old == new:
                    continue

                # replace the linear bias
                adj[new] = {new: adj[old].pop(old)}

                # copy the quadratic biases
                for v in adj[old]:
                    adj[new][v] = adj[v][new] = adj[v].pop(old)

                # remove the old adj for old
                del adj[old]

    def relabel_variables_as_integers(self) -> Mapping[int, Variable]:
        mapping = dict((v, i) for i, v in enumerate(self.variables) if i != v)
        self.relabel_variables(mapping)
        return dict((i, v) for v, i in mapping.items())

    def remove_variable(self, v: Optional[Variable] = None) -> Variable:
        if v is None:
            try:
                v, Nv = self._adj.popitem()
            except KeyError:
                raise ValueError("cannot pop from an empty model") from None
        else:
            try:
                Nv = self._adj.pop(v)
            except KeyError:
                raise ValueError(f"unknown variable {v!r}") from None

        for u in Nv:
            if u != v:
                self._adj[u].pop(v)

        return v

    def resize(self, n: int):
        while n > self.num_variables():
            self.add_variable()
        while n < self.num_variables():
            self.remove_variable()

    def set_linear(self, v: Variable, bias: Any):
        self._adj.setdefault(v, dict())[v] = bias

    def set_quadratic(self, u: Variable, v: Variable, bias: Any):
        if u == v:
            raise ValueError(f"{u!r} cannot have an interaction with itself")
        self.add_variable(u)
        self.add_variable(v)
        self._adj[u][v] = self._adj[v][u] = bias

    def to_numpy_vectors(self, *args, **kwargs):
        raise NotImplementedError  # defer to the caller

    def update(self, *args, **kwargs):
        raise NotImplementedError  # defer to the caller

    def vartype(self, v: Optional[Variable] = None) -> Vartype:
        return self._vartype



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import warnings

from dimod.vartypes import Vartype


def fix_variables(bqm, sampling_mode=True):
    """Determine assignments for some variables of a binary quadratic model.

    Roof duality finds a lower bound for the minimum of a quadratic polynomial. It
    can also find minimizing assignments for some of the polynomial's variables;
    these fixed variables take the same values in all optimal solutions [BHT]_ [BH]_.
    A quadratic pseudo-Boolean function can be represented as a network to find
    the lower bound through network-flow computations. `fix_variables` uses maximum
    flow in the implication network to correctly fix variables. Consequently, you can
    find an assignment for the remaining variables that attains the optimal value.

    Args:
        bqm (:obj:`.BinaryQuadraticModel`)
            A binary quadratic model.

        sampling_mode (bool, optional, default=True):
            In sampling mode, only roof-duality is used. When `sampling_mode` is false, strongly
            connected components are used to fix more variables, but in some optimal solutions
            these variables may take different values.

    Returns:
        dict: Variable assignments for some variables of the specified binary quadratic model.

    .. [BHT] Boros, E., P.L. Hammer, G. Tavares. Preprocessing of Unconstraint Quadratic Binary
        Optimization. Rutcor Research Report 10-2006, April, 2006.

    .. [BH] Boros, E., P.L. Hammer. Pseudo-Boolean optimization. Discrete Applied Mathematics 123,
        (2002), pp. 155-225

    """
    try:
        from dwave.preprocessing import roof_duality
    except ImportError:
        raise TypeError("you must install dwave-preprocessing to use this function") from None

    warnings.warn("fix_variables() is deprecated and will be removed in dimod 0.11.0, "
                  "please install dwave-preprocessing and use dwave.preprocessing.roof_duality "
                  "instead", DeprecationWarning, stacklevel=2)

    _, mapping = roof_duality(bqm, strict=sampling_mode)
    return mapping



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import division

import itertools
import collections.abc as abc

from numbers import Number

import numpy as np

from dimod.decorators import vartype_argument
from dimod.sampleset import as_samples
from dimod.utilities import iter_safe_relabels
from dimod.vartypes import Vartype

__all__ = ['BinaryPolynomial']


def asfrozenset(term):
    """Convert to frozenset if it is not already"""
    return term if isinstance(term, frozenset) else frozenset(term)


class BinaryPolynomial(abc.MutableMapping):
    """A polynomial with binary variables and real-valued coefficients.

    Args:
        poly (mapping/iterable):
            Polynomial as a mapping of form {term: bias, ...}, where `term` is
            a collection of variables and `bias` the associated bias. It can also
            be an iterable of 2-tuples (term, bias).

        vartype (:class:`.Vartype`/str/set):
            Variable type for the binary quadratic model. Accepted input values:

            * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

    Attributes:
        degree (int):
            Degree of the polynomial.

        variables (set):
            Variables of the polynomial.

        vartype (:class:`.Vartype`):
            One of :class:`~dimod.Vartype.SPIN` or :class:`~dimod.Vartype.BINARY`.

    Examples:

        Binary polynomials can be constructed in many different ways. The
        following are all equivalent

        >>> poly = dimod.BinaryPolynomial({'a': -1, 'ab': 1}, dimod.SPIN)
        >>> poly = dimod.BinaryPolynomial({('a',): -1, ('a', 'b'): 1}, dimod.SPIN)
        >>> poly = dimod.BinaryPolynomial([('a', -1), (('a', 'b'), 1)], dimod.SPIN)
        >>> poly = dimod.BinaryPolynomial({'a': -1, 'ab': .5, 'ba': .5}, dimod.SPIN)

        Binary polynomials act a mutable mappings but the terms can be accessed with
        any sequence.

        >>> poly = dimod.BinaryPolynomial({'a': -1, 'ab': 1}, dimod.BINARY)
        >>> poly['ab']
        1
        >>> poly['ba']
        1
        >>> poly[{'a', 'b'}]
        1
        >>> poly[('a', 'b')]
        1
        >>> poly['cd'] = 4
        >>> poly['dc']
        4

    """
    @vartype_argument('vartype')
    def __init__(self, poly, vartype):
        if isinstance(poly, abc.Mapping):
            poly = poly.items()

        # we need to aggregate the repeated terms
        self._terms = terms = {}
        for term, bias in poly:

            fsterm = asfrozenset(term)

            # when SPIN-valued, s^2 == 1, so we need to handle that case
            # in BINARY, x^2 == x
            if len(fsterm) < len(term) and vartype is Vartype.SPIN:
                new = set()
                term = tuple(term)  # make sure it has .count
                for v in fsterm:
                    if term.count(v) % 2:
                        new.add(v)
                fsterm = frozenset(new)

            if fsterm in terms:
                terms[fsterm] += bias
            else:
                terms[fsterm] = bias

        self.vartype = vartype

    def __contains__(self, term):
        return asfrozenset(term) in self._terms

    def __delitem__(self, term):
        del self._terms[asfrozenset(term)]

    def __eq__(self, other):
        if not isinstance(other, BinaryPolynomial):
            try:
                other = type(self)(other, self.vartype)
            except Exception:
                # not a polynomial
                return False

        self_terms = self._terms
        other_terms = other._terms

        return (
            self.vartype == other.vartype
            and all(
                (not bias or other_terms.get(term, 0.) == bias)
                for term, bias in self.items()
            )
            and all(
                (not bias or self_terms.get(term, 0.) == bias)
                for term, bias in other.items()
            )
        )

    def __ne__(self, other):
        return not (self == other)

    def __getitem__(self, term):
        return self._terms[asfrozenset(term)]

    def __iter__(self):
        return iter(self._terms)

    def __len__(self):
        return len(self._terms)

    def __setitem__(self, term, bias):
        self._terms[asfrozenset(term)] = bias

    def __repr__(self):
        return '{!s}({!r}, {!r})'.format(self.__class__.__name__,
                                         self._terms, self.vartype.name)

    @property
    def variables(self):
        """Variables of the polynomial."""
        return set().union(*self._terms)

    @property
    def degree(self):
        """Degree of the polynomial."""
        if len(self) == 0:
            return 0
        return max(map(len, self._terms))

    def copy(self):
        """Create a shallow copy."""
        return type(self)(self, self.vartype)

    def energy(self, sample_like, dtype=float):
        """Energy of the given sample.

        Args:
            sample_like (samples_like):
                A raw sample. `sample-like` is an extension of
                NumPy's array_like structure. See :func:`.as_samples`.

            dtype (:class:`numpy.dtype`, optional):
                Data type of the returned energy. Defaults to float.

        Returns:
            The energy.

        Examples:
            >>> poly = dimod.BinaryPolynomial({('x', 'y'): -1, ('x', 'y', 'z'): 0.5},
            ...                               'BINARY')
            >>> poly.energy({'x': 1, 'y': 1, 'z': 0})
            -1.0
        """
        energy, = self.energies(sample_like, dtype=dtype)
        return energy

    def energies(self, samples_like, dtype=float):
        """Energies of the given samples.

        Args:
            samples_like (samples_like):
                A collection of raw samples. `samples-like` is an extension of
                NumPy's array_like structure. See :func:`.as_samples`.

            dtype (:class:`numpy.dtype`, optional):
                Data type of the returned energies. Defaults to float.

        Returns:
            :obj:`numpy.ndarray`: The energies.

        Examples:
            >>> poly = dimod.BinaryPolynomial({('x', 'y'): -1, ('x', 'y', 'z'): 0.5},
            ...                               'BINARY')
            >>> samples = [{'x': 1, 'y': 1, 'z': 0},
            ...            {'x': 1, 'y': 1, 'z': 1}]
            >>> poly.energies(samples)
            array([-1. , -0.5])
        """
        samples, labels = as_samples(samples_like)
        if labels:
            idx, label = zip(*enumerate(labels))
            labeldict = dict(zip(label, idx))
        else:
            labeldict = {}

        num_samples = samples.shape[0]

        energies = np.zeros(num_samples, dtype=dtype)
        for term, bias in self.items():
            if len(term) == 0:
                energies += bias
            else:
                energies += np.prod([samples[:, labeldict[v]] for v in term], axis=0) * bias

        return energies

    def relabel_variables(self, mapping, inplace=True):
        """Relabel variables of a binary polynomial as specified by mapping.

        Args:
            mapping (dict):
                Dict mapping current variable labels to new ones. If an
                incomplete mapping is provided, unmapped variables retain their
                current labels.

            inplace (bool, optional, default=True):
                If True, the binary polynomial is updated in-place; otherwise, a
                new binary polynomial is returned.

        Returns:
            :class:`.BinaryPolynomial`: A binary polynomial with the variables
            relabeled. If `inplace` is set to True, returns itself.

        Examples:
            >>> poly = dimod.BinaryPolynomial({('x', 'y'): -1, ('x', 'y', 'z'): 0.5},
            ...                               'BINARY')
            >>> poly2 = poly.relabel_variables({'x': 'k', 'y': 'l'}, inplace=False)
            >>> 'k' in poly2.variables and 'z' in poly2.variables
            True
        """
        if not inplace:
            return self.copy().relabel_variables(mapping, inplace=True)

        for submap in iter_safe_relabels(mapping, self.variables):

            for oldterm, bias in list(self.items()):
                newterm = frozenset((submap.get(v, v) for v in oldterm))

                if newterm != oldterm:
                    self[newterm] = bias
                    del self[oldterm]

        return self

    def normalize(self, bias_range=1, poly_range=None, ignored_terms=None):
        """Normalize biases of the binary polynomial to fit the provided range(s).

        If ``poly_range`` is provided, ``bias_range`` is used as the range for
        linear biases only while ``poly_range`` is used for other biases.

        Args:
            bias_range (number/pair):
                Value/range by which to normalize all biases, or, if ``poly_range``
                is provided, just the linear biases.

            poly_range (number/pair, optional):
                Value/range by which to normalize higher-order biases.

            ignored_terms (iterable, optional):
                Biases associated with these terms are not scaled.

        """

        def parse_range(r):
            if isinstance(r, Number):
                return -abs(r), abs(r)
            return r

        if ignored_terms is None:
            ignored_terms = set()
        else:
            ignored_terms = {asfrozenset(term) for term in ignored_terms}

        if poly_range is None:
            linear_range, poly_range = bias_range, bias_range
        else:
            linear_range = bias_range

        lin_range, poly_range = map(parse_range, (linear_range, poly_range))

        # determine the current ranges for linear, higherorder
        lmin = lmax = 0
        pmin = pmax = 0
        for term, bias in self.items():

            if term in ignored_terms:
                # we don't use the ignored terms to calculate the scaling
                continue

            if len(term) == 1:
                lmin = min(bias, lmin)
                lmax = max(bias, lmax)
            elif len(term) > 1:
                pmin = min(bias, pmin)
                pmax = max(bias, pmax)

        inv_scalar = max(lmin / lin_range[0], lmax / lin_range[1],
                         pmin / poly_range[0], pmax / poly_range[1])

        if inv_scalar != 0:
            self.scale(1 / inv_scalar, ignored_terms=ignored_terms)

    def scale(self, scalar, ignored_terms=None):
        """Multiply the polynomial by the given scalar.

        Args:
            scalar (number):
                Value to multiply the polynomial by.

            ignored_terms (iterable, optional):
                Biases associated with these terms are not scaled.

        """

        if ignored_terms is None:
            ignored_terms = set()
        else:
            ignored_terms = {asfrozenset(term) for term in ignored_terms}

        for term in self:
            if term not in ignored_terms:
                self[term] *= scalar

    @classmethod
    def from_hising(cls, h, J, offset=None):
        """Construct a binary polynomial from a higher-order Ising problem.

        Args:
            h (dict):
                The linear biases.

            J (dict):
                The higher-order biases.

            offset (optional, default=0.0):
                Constant offset applied to the model.

        Returns:
            :obj:`.BinaryPolynomial`

        Examples:
            >>> poly = dimod.BinaryPolynomial.from_hising({'a': 2}, {('a', 'b'): -1}, 0)
            >>> poly.degree
            2

        """
        poly = {(k,): v for k, v in h.items()}
        poly.update(J)
        if offset is not None:
            poly[frozenset([])] = offset
        return cls(poly, Vartype.SPIN)

    def to_hising(self):
        """Construct a higher-order Ising problem from a binary polynomial.

        Returns:
            tuple: A 3-tuple of the form (`h`, `J`, `offset`) where `h` includes
            the linear biases, `J` has the higher-order biases and `offset` is
            the linear offset.

        Examples:
            >>> poly = dimod.BinaryPolynomial({'a': -1, ('a', 'b'): 1, ('a', 'b', 'c'): -1}, dimod.SPIN)
            >>> h, J, off = poly.to_hising()
            >>> h
            {'a': -1}

        """
        if self.vartype is Vartype.BINARY:
            return self.to_spin().to_hising()

        h = {}
        J = {}
        offset = 0
        for term, bias in self.items():
            if len(term) == 0:
                offset += bias
            elif len(term) == 1:
                v, = term
                h[v] = bias
            else:
                J[tuple(term)] = bias

        return h, J, offset

    @classmethod
    def from_hubo(cls, H, offset=None):
        """Construct a binary polynomial from a higher-order unconstrained
        binary optimization (HUBO) problem.

        Args:
            H (dict):
                Coefficients of a higher-order unconstrained binary optimization
                (HUBO) model.

        Returns:
            :obj:`.BinaryPolynomial`

        Examples:
            >>> poly = dimod.BinaryPolynomial.from_hubo({('a', 'b', 'c'): -1})
            >>> poly.degree
            3

        """
        poly = cls(H, Vartype.BINARY)
        if offset is not None:
            poly[()] = poly.get((), 0) + offset
        return poly

    def to_hubo(self):
        """Construct a higher-order unconstrained binary optimization (HUBO)
        problem from a binary polynomial.

        Returns:
            tuple: A 2-tuple of the form (`H`, `offset`) where `H` is the HUBO
            and `offset` is the linear offset.

        Examples:
            >>> poly = dimod.BinaryPolynomial({'a': -1, ('a', 'b', 'c'): -1}, dimod.BINARY)
            >>> H, off = poly.to_hubo()

        """
        if self.vartype is Vartype.SPIN:
            return self.to_binary().to_hubo()

        H = {tuple(term): bias for term, bias in self.items() if term}
        offset = self[tuple()] if tuple() in self else 0
        return H, offset

    def to_binary(self, copy=False):
        """Return a binary polynomial over `{0, 1}` variables.

        Args:
            copy (optional, default=False):
                If True, the returned polynomial is always a copy. Otherwise,
                if the polynomial is binary-valued already it returns itself.

        Returns:
            :obj:`.BinaryPolynomial`

        """
        if self.vartype is Vartype.BINARY:
            if copy:
                return self.copy()
            else:
                return self

        new = BinaryPolynomial({}, Vartype.BINARY)

        # s = 2x - 1
        for term, bias in self.items():
            for t in map(frozenset, powerset(term)):
                newbias = bias * 2**len(t) * (-1)**(len(term) - len(t))

                if t in new:
                    new[t] += newbias
                else:
                    new[t] = newbias

        return new

    def to_spin(self, copy=False):
        """Return a binary polynomial over `{-1, +1}` variables.

        Args:
            copy (optional, default=False):
                If True, the returned polynomial is always a copy. Otherwise,
                if the polynomial is spin-valued already it returns itself.

        Returns:
            :obj:`.BinaryPolynomial`

        """
        if self.vartype is Vartype.SPIN:
            if copy:
                return self.copy()
            else:
                return self

        new = BinaryPolynomial({}, Vartype.SPIN)

        # x = (s + 1) / 2
        for term, bias in self.items():
            newbias = bias / (2**len(term))

            for t in map(frozenset, powerset(term)):
                if t in new:
                    new[t] += newbias
                else:
                    new[t] = newbias

        return new


def powerset(iterable):
    return itertools.chain.from_iterable(itertools.combinations(iterable, r)
                                         for r in range(len(iterable)+1))



# distutils: language = c++
# cython: language_level=3

# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport cython
cimport numpy as np

from dimod.libcpp cimport cppVartype

__all__ = [
    'as_numpy_float',
    'coo_sort',
    'Integer',
    'SignedInteger',
    'UnsignedInteger',
    ]

cdef fused SignedInteger:
    np.int8_t
    np.int16_t
    np.int32_t
    np.int64_t

cdef fused UnsignedInteger:
    np.uint8_t
    np.uint16_t
    np.uint32_t
    np.uint64_t

cdef fused Integer:
    SignedInteger
    UnsignedInteger

cdef fused ConstInteger:
    const short
    const int
    const long long
    const unsigned short
    const unsigned int
    const unsigned long long

ctypedef fused Numeric:
    signed char
    cython.integral  # short, int, long
    cython.floating  # float, double

# cython doesn't like const Numeric
ctypedef fused ConstNumeric:
    const signed char
    const signed short
    const signed int
    const signed long long
    const float
    const double


cdef object as_numpy_float(cython.floating)


cpdef Py_ssize_t coo_sort(Integer[:], Integer[:], cython.floating[:]) except -1

cdef cppVartype cppvartype(object) except? cppVartype.SPIN



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Enumeration of valid variable types for binary quadratic models.

Examples:

    :class:`.Vartype` is an :py:class:`~enum.Enum`. Each vartype has a value and
    a name.

    >>> vartype = dimod.SPIN
    >>> vartype.name
    'SPIN'
    >>> vartype.value == {-1, +1}
    True

    >>> vartype = dimod.BINARY
    >>> vartype.name
    'BINARY'
    >>> vartype.value == {0, 1}
    True

    The :func:`.as_vartype` function allows the user to provide several
    convenient forms.

    >>> from dimod import as_vartype

    >>> as_vartype(dimod.SPIN) is dimod.SPIN
    True
    >>> as_vartype('SPIN') is dimod.SPIN
    True
    >>> as_vartype({-1, 1}) is dimod.SPIN
    True

    >>> as_vartype(dimod.BINARY) is dimod.BINARY
    True
    >>> as_vartype('BINARY') is dimod.BINARY
    True
    >>> as_vartype({0, 1}) is dimod.BINARY
    True

"""
import enum
import typing
import warnings

from collections.abc import Container, Iterable

__all__ = ['as_vartype',
           'Vartype', 'ExtendedVartype',
           'SPIN', 'BINARY', 'DISCRETE', 'INTEGER',
           ]


class Integers(Container):
    """Container for testing integer membership."""
    def __contains__(self, item):
        try:
            return int(item) == item
        except TypeError:
            return False

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self


integers = Integers()


class Vartype(enum.Enum):
    """An :py:class:`~enum.Enum` over the types of variables for quadratic models.

    Attributes:
        SPIN: Vartype for spin-valued binary quadratic models and variables of
           quadratic models that have values that are either -1 or 1.
        BINARY: Vartype for binary quadratic models and variables of quadratic
           models that have values that are either 0 or 1.
        INTEGER: Vartype for variables in quadratic models
            that have values of type int.
    """
    SPIN = frozenset({-1, 1})
    BINARY = frozenset({0, 1})
    INTEGER = integers
    DISCRETE = integers  # alias for INTEGER


# Deprecated alias
ExtendedVartype = Vartype


SPIN = Vartype.SPIN
BINARY = Vartype.BINARY
INTEGER = Vartype.INTEGER
DISCRETE = Vartype.DISCRETE


VartypeLike = typing.Union[Vartype, str, frozenset]


def _vartype_miss(vartype, extended):
    if extended:
        candidates = ("Vartype.SPIN, 'SPIN', {-1, 1}, "
                      "Vartype.BINARY, 'BINARY', {0, 1},"
                      "Vartype.INTEGER, or 'INTEGER'")
    else:
        candidates = ("Vartype.SPIN, 'SPIN', {-1, 1}, "
                      "Vartype.BINARY, 'BINARY', or {0, 1}")
    return TypeError(f"expected input vartype to be one of: {candidates!s}; "
                     f"received {vartype!r}.")


# In the future we may wish to make extended default to True. However, for
# now even raising a PendingDeprecationWarning would be a big change.
def as_vartype(vartype: VartypeLike, extended: bool = False) -> Vartype:
    """Cast various inputs to a valid vartype object.

    Args:
        vartype (:class:`.Vartype`/str/set):
            Variable type. Accepted input values:

            * :class:`~dimod.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            * :class:`~dimod.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

        extended (bool, optional, default=False):
            If `True`, vartype can also be:

            * :class:`~dimod.Vartype.INTEGER`, ``'INTEGER'``
            * :class:`~dimod.Vartype.DISCRETE`, ``'DISCRETE'``


    Returns:
        :class:`.Vartype`: Either :class:`~dimod.Vartype.SPIN` or
        :class:`~dimod.Vartype.BINARY`. If `extended` is True, can also
        be :class:`~dimod.Vartype.INTEGER`.

    See also:
        :func:`~dimod.decorators.vartype_argument`

    """
    # can be done with singledispatch, but this function needs to be
    # performant and if branches are faster
    if isinstance(vartype, Vartype):
        pass  # already what we want
    elif isinstance(vartype, str):
        try:
            vartype = Vartype[vartype]
        except KeyError:
            raise _vartype_miss(vartype, extended) from None
    elif isinstance(vartype, frozenset):
        try:
            vartype = Vartype(vartype)
        except ValueError:
            raise _vartype_miss(vartype, extended) from None
    elif isinstance(vartype, Iterable):  # not frozenset or str
        try:
            vartype = Vartype(frozenset(vartype))
        except ValueError:
            raise _vartype_miss(vartype, extended) from None
    else:
        raise _vartype_miss(vartype, extended)

    if not extended and vartype != SPIN and vartype != BINARY:
        raise _vartype_miss(vartype, extended)

    return vartype



# distutils: language = c++
# cython: language_level=3

# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport cython

from dimod.cyutilities cimport ConstNumeric
from dimod.cyvariables cimport cyVariables
from dimod.quadratic.cyqm.cyqm_base cimport cyQMBase
from dimod.libcpp cimport cppQuadraticModel, cppVartype


cdef class cyQM_template(cyQMBase):
    cdef cppQuadraticModel[bias_type, index_type] cppqm

    cdef readonly object dtype
    cdef readonly object index_dtype
    cdef readonly cyVariables variables

    cdef void _add_linear(self, Py_ssize_t, bias_type)
    cdef np.float64_t[::1] _energies(self, ConstNumeric[:, ::1] samples, object labels)
    cdef void _set_linear(self, Py_ssize_t, bias_type)
    cdef cppVartype cppvartype(self, object) except? cppVartype.SPIN
    cdef const cppQuadraticModel[bias_type, index_type]* data(self)
    cpdef bint is_linear(self)
    cpdef void scale(self, bias_type)
    cpdef Py_ssize_t nbytes(self, bint capacity=*)
    cpdef Py_ssize_t num_interactions(self)
    cpdef Py_ssize_t num_variables(self)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A composite that clips problem variables below and above threshold. if lower
and upper bounds is not given it does nothing.

"""

try:
    from dwave.preprocessing import ClipComposite as _ClipComposite
except ImportError:
    from dimod.reference.composites._preprocessing import NotFound as _ClipComposite

from dimod.reference.composites._preprocessing import DeprecatedToPreprocessing


__all__ = ['ClipComposite']


class ClipComposite(DeprecatedToPreprocessing, _ClipComposite):
    pass


ClipComposite.__doc__ = _ClipComposite.__doc__
